"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging

import numpy as np
from ase.stress import voigt_6_to_full_3x3_stress


def uniform_atoms_lengths(atoms_lens) -> bool:
    # If all of the structures have the same number of atoms, it's really hard to know
    # whether the entries are intensive or extensive, and whether
    # some of the entries are per-atom or not
    return len(set(atoms_lens)) == 1


def target_constant_shape(atoms_lens, target_samples) -> bool:
    # Given a bunch of atoms lengths, and the corresponding samples for the target,
    # determine whether the shape is always the same regardless of atom size
    return len({sample.shape for sample in target_samples}) == 1


def target_per_atom(atoms_lens, target_samples) -> bool:
    # Given a bunch of atoms lengths, and the corresponding samples for the target,
    # determine whether the target is per-atom (first dimension == # atoms, others constant)

    # If a sample target is just a number/float/etc, it can't be per-atom
    if len(np.array(target_samples[0]).shape) == 0:
        return False

    first_dim_proportional = all(
        np.array(sample).shape[0] == alen
        for alen, sample in zip(atoms_lens, target_samples)
    )

    if len(np.array(target_samples[0]).shape) == 1:
        other_dim_constant = True
    else:
        other_dim_constant = (
            len({np.array(sample).shape[1:] for sample in target_samples}) == 1
        )

    return bool(first_dim_proportional and other_dim_constant)


def target_extensive(atoms_lens, target_samples, threshold: float = 0.2):
    # Guess whether a property is intensive or extensive.
    # We guess by checking whether standard deviation of the per-atom
    # properties capture >20% of the variation in the property
    # Of course, with a small amount of data!

    # If the targets are all the same shapes, we shouldn't be asking if the property
    # is intensive or extensive!
    assert target_constant_shape(
        atoms_lens, target_samples
    ), "The shapes of this target are not constant!"

    # Get the per-atom normalized properties
    try:
        compiled_target_array = np.array(
            [sample / atom_len for sample, atom_len in zip(atoms_lens, target_samples)]
        )
    except TypeError:
        return False

    # Calculate the normalized standard deviation of each element in the property output
    target_samples_mean = np.mean(compiled_target_array, axis=0)
    target_samples_normalized = compiled_target_array / target_samples_mean

    # If there's not much variation in the per-atom normalized properties,
    # guess extensive!
    extensive_guess = target_samples_normalized.std(axis=0) < (
        threshold * target_samples_normalized.mean(axis=0)
    )
    if extensive_guess.shape == ():
        return extensive_guess
    return bool(
        (
            target_samples_normalized.std(axis=0)
            < threshold * target_samples_normalized.mean(axis=0)
        ).all()
    )


def guess_target_metadata(atoms_len, target_samples):
    example_array = np.array(target_samples[0])
    if example_array.dtype == object or example_array.dtype == str:
        return {
            "shape": None,
            "type": "unknown",
            "extensive": None,
            "units": "unknown",
            "comment": "Guessed property metadata. The property didn't seem to be a numpy array with any numeric type, so we dob't know what to do.",
        }
    elif target_constant_shape(atoms_len, target_samples):
        target_shape = np.array(target_samples[0]).shape

        if uniform_atoms_lengths(atoms_len):
            if atoms_len[0] > 3 and target_per_atom(atoms_len, target_samples):
                target_shape = list(target_samples[0].shape)
                target_shape[0] = "N"
                return {
                    "shape": tuple(target_shape),
                    "type": "per-atom",
                    "extensive": True,
                    "units": "unknown",
                    "comment": "Guessed property metadata. Because all the sampled atoms are the same length, we can't really know if it is per-atom or per-frame, but the first dimension happens to match the number of atoms.",
                }
            else:
                return {
                    "shape": tuple(target_shape),
                    "type": "per-image",
                    "extensive": True,
                    "units": "unknown",
                    "comment": "Guessed property metadata. Because all the sampled atoms are the same length, we can't know if this is intensive of extensive, or per-image or per-frame",
                }

        elif target_extensive(atoms_len, target_samples):
            return {
                "shape": tuple(target_shape),
                "type": "per-image",
                "extensive": True,
                "comment": "Guessed property metadata. It appears to be extensive based on a quick correlation with atom sizes",
            }
        else:
            return {
                "shape": tuple(target_shape),
                "type": "per-image",
                "extensive": False,
                "units": "unknown",
                "comment": "Guess property metadata. It appears to be intensive based on a quick correlation with atom sizes.",
            }
    elif target_per_atom(atoms_len, target_samples):
        target_shape = list(target_samples[0].shape)[1:]
        return {
            "shape": tuple(target_shape),
            "type": "per-atom",
            "extensive": True,
            "units": "unknown",
            "comment": "Guessed property metadata. It appears to be a per-atom property.",
        }
    else:
        return {
            "shape": None,
            "type": "unknown",
            "extensive": None,
            "units": "unknown",
            "comment": "Guessed property metadata. The property was variable across different samples and didn't seem to be a per-atom property",
        }


def guess_property_metadata(atoms_list):
    atoms = atoms_list[0]
    atoms_len = [len(atoms) for atoms in atoms_list]

    targets = {}

    if hasattr(atoms, "info"):
        for key in atoms.info:
            # Grab the property samples from the list of atoms
            target_samples = [np.array(atoms.info[key]) for atoms in atoms_list]

            # Guess the metadata
            targets[f"info.{key}"] = guess_target_metadata(atoms_len, target_samples)

            # Log a warning so the user knows what's happening
            logging.warning(
                f'Guessed metadata for atoms.info["{key}"]: {targets[f"info.{key}"]!s}'
            )
    if hasattr(atoms, "calc") and atoms.calc is not None:
        for key in atoms.calc.results:
            # Grab the property samples from the list of atoms
            target_samples = [np.array(atoms.calc.results[key]) for atoms in atoms_list]

            # stress needs to be handled separately in case it was saved in voigt (6, ) notation
            # atoms2graphs will always request voigt=False so turn it into full 3x3
            if key == "stress":
                target_samples = [
                    voigt_6_to_full_3x3_stress(sample)
                    if sample.shape != (3, 3)
                    else sample
                    for sample in target_samples
                ]

            # Guess the metadata
            targets[f"{key}"] = guess_target_metadata(atoms_len, target_samples)

            # Log a warning so the user knows what's happening
            logging.warning(
                f'Guessed metadata for ASE calculator property ["{key}"]: {targets[key]!s}'
            )

    return targets



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Union

import torch
import torch.nn as nn

from .scale_factor import ScaleFactor

ScaleDict = Union[dict[str, float], dict[str, torch.Tensor]]


def _load_scale_dict(scale_file: str | ScaleDict | None):
    """
    Loads scale factors from either:
    - a JSON file mapping scale factor names to scale values
    - a python dictionary pickled object (loaded using `torch.load`) mapping scale factor names to scale values
    - a dictionary mapping scale factor names to scale values
    """
    if not scale_file:
        return None

    if isinstance(scale_file, dict):
        if not scale_file:
            logging.warning("Empty scale dictionary provided to model.")
        return scale_file

    path = Path(scale_file)
    if not path.exists():
        raise ValueError(f"Scale file {path} does not exist.")

    scale_dict: ScaleDict | None = None
    if path.suffix == ".pt":
        scale_dict = torch.load(path)
    elif path.suffix == ".json":
        with open(path) as f:
            scale_dict = json.load(f)

        if isinstance(scale_dict, dict):
            # old json scale factors have a comment field that has the model name
            scale_dict.pop("comment", None)
    else:
        raise ValueError(f"Unsupported scale file extension: {path.suffix}")

    if not scale_dict:
        return None

    return scale_dict


def load_scales_compat(module: nn.Module, scale_file: str | ScaleDict | None) -> None:
    scale_dict = _load_scale_dict(scale_file)
    if not scale_dict:
        return

    scale_factors = {
        module.name or name: (module, name)
        for name, module in module.named_modules()
        if isinstance(module, ScaleFactor)
    }
    logging.debug(
        f"Found the following scale factors: {[(k, name) for k, (_, name) in scale_factors.items()]}"
    )
    missing_keys = set(scale_factors.keys()) - set(scale_dict.keys())
    if len(missing_keys) > 0:
        raise ValueError(
            "Failed to load scaling values. Missing entries for,",
            missing_keys,
            "\nHave",
            scale_dict.keys(),
        )
    for name, scale in scale_dict.items():
        if name not in scale_factors:
            logging.warning(f"Scale factor {name} not found in model")
            continue
        scale_module, module_name = scale_factors[name]
        logging.debug(f"Loading scale factor {scale} for ({name} => {module_name})")
        scale_module.set_(scale)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
from typing import TypeVar

import torch

from fairchem.core.datasets.atomic_data import AtomicData, atomicdata_list_to_batch

T_co = TypeVar("T_co", covariant=True)


def data_list_collater(
    data_list: list[AtomicData], otf_graph: bool = False, to_dict: bool = False
) -> AtomicData | dict[str, torch.Tensor]:
    # batch = Batch.from_data_list(data_list)
    batch = atomicdata_list_to_batch(data_list)

    if not otf_graph:
        try:
            n_neighbors = []
            for _, data in enumerate(data_list):
                n_index = data.edge_index[1, :]
                n_neighbors.append(n_index.shape[0])
            batch.neighbors = torch.tensor(n_neighbors)
        except (NotImplementedError, TypeError):
            logging.warning("No edge index information, set otf_graph=True")

    if to_dict:
        batch = dict(batch.items())

    return batch



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
from abc import ABCMeta
from functools import cached_property
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)

import numpy as np
import torch
from torch import randperm
from torch.utils.data import Dataset
from torch.utils.data import Subset as Subset_

from fairchem.core.common.registry import registry

if TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import ArrayLike


T_co = TypeVar("T_co", covariant=True)


class UnsupportedDatasetError(ValueError):
    pass


class BaseDataset(Dataset[T_co], metaclass=ABCMeta):
    """Base Dataset class for all OCP datasets."""

    def __init__(self, config: dict):
        """Initialize

        Args:
            config (dict): dataset configuration
        """
        self.config = config
        self.paths = []

        if "src" in self.config:
            if isinstance(config["src"], str):
                self.paths = [Path(self.config["src"])]
            else:
                self.paths = tuple(Path(path) for path in sorted(config["src"]))

    def __len__(self) -> int:
        return self.num_samples

    def metadata_hasattr(self, attr) -> bool:
        return attr in self._metadata

    @cached_property
    def indices(self):
        return np.arange(self.num_samples, dtype=int)

    @cached_property
    def _metadata(self) -> dict[str, ArrayLike]:
        # logic to read metadata file here
        metadata_npzs = []
        if self.config.get("metadata_path", None) is not None:
            metadata_npzs.append(
                np.load(self.config["metadata_path"], allow_pickle=True)
            )

        else:
            for path in self.paths:
                if path.is_file():
                    metadata_file = path.parent / "metadata.npz"
                else:
                    metadata_file = path / "metadata.npz"
                if metadata_file.is_file():
                    metadata_npzs.append(np.load(metadata_file, allow_pickle=True))

        if len(metadata_npzs) == 0:
            logging.warning(
                f"Could not find dataset metadata.npz files in '{self.paths}'"
            )
            return {}

        metadata = {
            field: np.concatenate([metadata[field] for metadata in metadata_npzs])
            for field in metadata_npzs[0]
        }

        assert np.issubdtype(
            metadata["natoms"].dtype, np.integer
        ), f"Metadata natoms must be an integer type! not {metadata['natoms'].dtype}"
        assert (
            metadata["natoms"].shape[0] == len(self)
        ), f"Loaded metadata size {metadata['natoms'].shape[0]} and dataset size {len(self)} mismatch."

        return metadata

    def get_metadata(self, attr, idx):
        if attr in self._metadata:
            metadata_attr = self._metadata[attr]
            if isinstance(idx, list):
                return [metadata_attr[_idx] for _idx in idx]
            return metadata_attr[idx]
        return None


class Subset(Subset_, BaseDataset):
    """A pytorch subset that also takes metadata if given."""

    def __init__(
        self,
        dataset: BaseDataset,
        indices: Sequence[int],
        metadata: dict[str, ArrayLike],
    ) -> None:
        super().__init__(dataset, indices)
        self.metadata = metadata
        self.indices = indices
        self.num_samples = len(indices)
        self.config = dataset.config

    @cached_property
    def _metadata(self) -> dict[str, ArrayLike]:
        return self.dataset._metadata

    def get_metadata(self, attr, idx):
        if isinstance(idx, list):
            return self.dataset.get_metadata(attr, [[self.indices[i] for i in idx]])
        return self.dataset.get_metadata(attr, self.indices[idx])


def create_dataset(config: dict[str, Any], split: str) -> Subset:
    """Create a dataset from a config dictionary

    Args:
        config (dict): dataset config dictionary
        split (str): name of split

    Returns:
        Subset: dataset subset class
    """
    # Initialize the dataset
    dataset_cls = registry.get_dataset_class(config.get("format", "lmdb"))
    assert issubclass(dataset_cls, Dataset), f"{dataset_cls} is not a Dataset"

    # remove information about other splits, only keep specified split
    # this may only work with the mt config not main config
    current_split_config = config.copy()
    if "splits" in current_split_config:
        current_split_config.pop("splits")
        current_split_config.update(config["splits"][split])

    seed = current_split_config.get("seed", 0)
    if split != "train":
        seed += (
            1  # if we use same dataset for train / val , make sure its diff sampling
        )

    g = torch.Generator()
    g.manual_seed(seed)

    dataset = dataset_cls(current_split_config)

    # Get indices of the dataset
    indices = dataset.indices
    max_atoms = current_split_config.get("max_atoms", None)
    if max_atoms is not None:
        if not dataset.metadata_hasattr("natoms"):
            raise ValueError("Cannot use max_atoms without dataset metadata")
        indices = indices[dataset.get_metadata("natoms", indices) <= max_atoms]

    for subset_to in current_split_config.get("subset_to", []):
        if not dataset.metadata_hasattr(subset_to["metadata_key"]):
            raise ValueError(
                f"Cannot use {subset_to} without dataset metadata key {subset_to['metadata_key']}"
            )
        rhv = subset_to["rhv"]
        if isinstance(rhv, str):
            with open(rhv) as f:
                rhv = f.read().splitlines()
                rhv = [int(x) for x in rhv]
        if subset_to["op"] == "abs_le":
            indices = indices[
                np.abs(dataset.get_metadata(subset_to["metadata_key"], indices)) <= rhv
            ]
        elif subset_to["op"] == "in":
            indices = indices[
                np.isin(dataset.get_metadata(subset_to["metadata_key"], indices), rhv)
            ]

    # Apply dataset level transforms
    # TODO is no_shuffle mutually exclusive though? or what is the purpose of no_shuffle?
    first_n = current_split_config.get("first_n")
    sample_n = current_split_config.get("sample_n")
    no_shuffle = current_split_config.get("no_shuffle")
    # this is true if at most one of the mutually exclusive arguments are set
    if sum(arg is not None for arg in (first_n, sample_n, no_shuffle)) > 1:
        raise ValueError(
            "sample_n, first_n, no_shuffle are mutually exclusive arguments. Only one can be provided."
        )
    if first_n is not None:
        max_index = first_n
    elif sample_n is not None:
        # shuffle by default, user can disable to optimize if they have confidence in dataset
        # shuffle all datasets by default to avoid biasing the sampling in concat dataset
        # TODO only shuffle if split is train
        max_index = sample_n
        indices = (
            indices
            if len(indices) == 1
            else indices[randperm(len(indices), generator=g)]
        )
    else:
        max_index = len(indices)
        indices = (
            indices
            if (no_shuffle or len(indices) == 1)
            else indices[randperm(len(indices), generator=g)]
        )

    if max_index > len(indices):
        msg = (
            f"Cannot take {max_index} data points from a dataset of only length {len(indices)}.\n"
            f"Make sure to set first_n or sample_n to a number =< the total samples in dataset."
        )
        if max_atoms is not None:
            msg = msg[:-1] + f"that are smaller than the given max_atoms {max_atoms}."
        raise ValueError(msg)

    indices = indices[:max_index]

    return Subset(dataset, indices, metadata=dataset._metadata)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import pandas as pd
from monty.dev import requires
from pymatviz.enums import Key

from fairchem.core.components.benchmark.benchmark_reducer import JsonDFReducer
from fairchem.core.components.calculate.kappa_runner import KappaRunner

if TYPE_CHECKING:
    from collections.abc import Sequence

try:
    from matbench_discovery.metrics import phonons

    mbd_installed = True
except ImportError:
    Kappa103_TARGET_DATA_PATH = None
    mbd_installed = False


@requires(mbd_installed, message="Requires `matbench_discovery` to be installed")
class Kappa103Reducer(JsonDFReducer):
    def __init__(
        self,
        benchmark_name: str,
        target_data_path: Optional[str] = None,
        target_data_keys: Sequence[str] | None = None,
        index_name: (
            str | None
        ) = "mp_id",  # bug in matbench-discovery on column name file content mismatch: mp_id vs Key.mat_id
    ):
        """
        Args:
            benchmark_name: Name of the benchmark, used for file naming
            target_data_path: Path to the target data JSON file
            index_name: Optional name of the column to use as index
            corrections: Optional correction class to apply to all entries
            max_error_threshold: Maximum allowed mean absolute formation energy per atom error threshold
        """
        index_name = index_name or str(Key.mat_id)
        super().__init__(benchmark_name, target_data_path, target_data_keys, index_name)

    @property
    def runner_type(self) -> type[KappaRunner]:
        """The runner type this reducer is associated with."""
        return KappaRunner

    def compute_metrics(self, results: pd.DataFrame, run_name: str) -> pd.DataFrame:
        """Compute Matbench discovery metrics for relaxed energy and structure predictions.

        Args:
            results: DataFrame containing prediction results with energy values
            run_name: Identifier for the current evaluation run

        Returns:
            DataFrame containing computed metrics for different material subsets
        """
        target_data = self.target_data.rename(columns={"mp_id": Key.mat_id})
        kappa_metrics = phonons.calc_kappa_metrics_from_dfs(results, target_data)
        metrics = {
            "kappa_sre": kappa_metrics[Key.sre].mean(),
            "kappa_srme": kappa_metrics[Key.srme].mean(),
        }
        return pd.DataFrame([metrics], index=[run_name])

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        pass

    def load_state(self, checkpoint_location: str | None) -> None:
        pass



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import argparse
import logging
from pathlib import Path
from urllib.request import urlretrieve


def fairchem_root():
    """Return the root directory of the installed fairchem-core package."""
    import fairchem.core

    return Path(fairchem.core.__file__).parent.parent


S3_ROOT = "https://dl.fbaipublicfiles.com/opencatalystproject/data/large_files/"

FILE_GROUPS = {
    "odac": [
        Path("configs/odac/s2ef/scaling_factors/painn.pt"),
        Path("src/fairchem/data/odac/force_field/data_w_oms.json"),
        Path(
            "src/fairchem/data/odac/promising_mof/promising_mof_features/JmolData.jar"
        ),
        Path(
            "src/fairchem/data/odac/promising_mof/promising_mof_energies/adsorption_energy.txt"
        ),
        Path("src/fairchem/data/odac/supercell_info.csv"),
    ],
    "oc": [Path("src/fairchem/data/oc/databases/pkls/bulks.pkl")],
    "adsorbml": [
        Path(
            "src/fairchem/applications/AdsorbML/adsorbml/2023_neurips_challenge/oc20dense_mapping.pkl"
        ),
        Path(
            "src/fairchem/applications/AdsorbML/adsorbml/2023_neurips_challenge/ml_relaxed_dft_targets.pkl"
        ),
    ],
    "cattsunami": [
        Path("tests/applications/cattsunami/tests/autoframe_inputs_dissociation.pkl"),
        Path("tests/applications/cattsunami/tests/autoframe_inputs_transfer.pkl"),
    ],
    "docs": [
        Path("docs/tutorials/NRR/NRR_example_bulks.pkl"),
        Path("docs/core/fine-tuning/supporting-information.json"),
        Path("docs/core/data.db"),
    ],
}


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file_group",
        type=str,
        help="Group of files to download",
        default="ALL",
        choices=["ALL", *list(FILE_GROUPS)],
    )
    return parser.parse_args()


def change_path_for_pypi(
    files_to_download: list[Path],
    par_dir: str,
    install_dir: Path,
    test_par_dir: Path | None,
) -> list[Path]:
    """
    Modify or exclude files from download if running in a PyPi-installed
    build.

    Installation of FAIR-Chem with PyPi does not include the entire
    directory structure of the fairchem repo. As such, files outside of
    `src` can't be downloaded and those in `src` should actually be in the
    `site-packages` directory. If a user wants these files, they must
    build from the git repo.

    If the tests have been separately downloaded (e.g. from the git repo),
    then we can download if we've been told where those tests have been
    downloaded to. Note that we can't divine that location from anything
    in fairchem.core because they would have to be somewhere "unexpected"
    since we've built with PyPi which shouldn't have tests at all.

    :param files_to_download: List of files to be downloaded
    :param par_dir: the parent directory of the PyPi build,
                    probably "site-packages"
    :param install_dir: path to where fairchem.core was installed
    :param test_par_dir: path to where tests have been downloaded
                         (not necessarily the same as install_dir)
    :return: modified list of files to be downloaded
    """
    new_files = []
    for file in files_to_download:
        # We check the top-level name of the file so we know if it has a
        # home in PyPi builds
        top_level_name = file.parts[0]
        if top_level_name == "tests" and test_par_dir is not None:
            new_files.append(test_par_dir / file)
        elif top_level_name == "src":
            # turn `src` into `site-packages` or whatever the correct name is
            new_files.append(install_dir / Path(str(file).replace("src", par_dir, 1)))
    return new_files


def download_file_group(file_group: str, test_par_dir: Path | None = None) -> None:
    """
    Download the given file group.

    :param file_group: Name of group of files to download
    :param test_par_dir: Parent directory where fairchem tests have been
                         downloaded
    """
    if file_group in FILE_GROUPS:
        files_to_download = FILE_GROUPS[file_group]
    elif file_group == "ALL":
        files_to_download = [item for group in FILE_GROUPS.values() for item in group]
    else:
        raise ValueError(
            f'Requested file group {file_group} not recognized. Please select one of {["ALL", *list(FILE_GROUPS)]}'
        )

    fc_root = fairchem_root()
    # The grandparent of fairchem_root is the install location,
    # this is the git repo top-level directory for dev builds, i.e. `fairchem`
    install_dir = fc_root.parent.parent
    # The parent of fairchem_root is `src` on git repo/dev builds and
    # usually `site-packages` for PyPi builds (to be safe, we don't assume
    # that name)
    fc_parent = str(fc_root.parent.name)
    if fc_parent != "src":
        files_to_download = change_path_for_pypi(
            files_to_download, fc_parent, install_dir, test_par_dir
        )
    else:
        files_to_download = [install_dir / file for file in files_to_download]

    missing_path = False
    for file in files_to_download:
        if not file.parent.exists():
            print(f"Cannot download {file}, the path does not exist.")
            missing_path = True
        elif not file.exists():
            print(f"Downloading {file}...")
            urlretrieve(S3_ROOT + file.name, file)
        else:
            print(f"{file} already exists")

    if missing_path:
        logging.error(
            "\n\nSome files could not be downloaded because their "
            "expected landing spot didn't exist. If you installed with "
            "PyPi perhaps additional packages like `fairchem-data-oc` "
            "are needed.\nIf all PyPi packages have been installed, you "
            "may need to install from the git repo to have the full "
            "fairchem contents."
        )


if __name__ == "__main__":
    args = parse_args()
    download_file_group(args.file_group)



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import torch
from torch import nn
from torch.nn import functional as F

if TYPE_CHECKING:
    from fairchem.core.models.escaip.configs import (
        GlobalConfigs,
        GraphNeuralNetworksConfigs,
        MolecularGraphConfigs,
        RegularizationConfigs,
    )
    from fairchem.core.models.escaip.custom_types import GraphAttentionData
from fairchem.core.models.escaip.modules.base_block import BaseGraphNeuralNetworkLayer
from fairchem.core.models.escaip.utils.nn_utils import (
    Activation,
    NormalizationType,
    get_feedforward,
    get_linear,
    get_normalization_layer,
)
from fairchem.core.models.escaip.utils.stochastic_depth import (
    SkipStochasticDepth,
    StochasticDepth,
)


class EfficientGraphAttentionBlock(nn.Module):
    """
    Efficient Graph Attention Block module.
    Ref: swin transformer
    """

    def __init__(
        self,
        global_cfg: GlobalConfigs,
        molecular_graph_cfg: MolecularGraphConfigs,
        gnn_cfg: GraphNeuralNetworksConfigs,
        reg_cfg: RegularizationConfigs,
        is_last: bool = False,
    ):
        super().__init__()

        self.backbone_dtype = (
            torch.float16 if global_cfg.use_fp16_backbone else torch.float32
        )

        # Graph attention
        self.graph_attention = EfficientGraphAttention(
            global_cfg=global_cfg,
            molecular_graph_cfg=molecular_graph_cfg,
            gnn_cfg=gnn_cfg,
            reg_cfg=reg_cfg,
        )

        # Feed forward network
        self.feedforward = FeedForwardNetwork(
            global_cfg=global_cfg,
            gnn_cfg=gnn_cfg,
            reg_cfg=reg_cfg,
            is_last=is_last,
        )

        # Normalization
        normalization = NormalizationType(reg_cfg.normalization)
        self.norm_attn_node = get_normalization_layer(normalization)(
            global_cfg.hidden_size, dtype=self.backbone_dtype
        )
        self.norm_attn_edge = get_normalization_layer(normalization)(
            global_cfg.hidden_size, dtype=self.backbone_dtype
        )
        self.norm_ffn_node = get_normalization_layer(normalization)(
            global_cfg.hidden_size, dtype=self.backbone_dtype
        )
        if not (
            (is_last) and (global_cfg.regress_forces and not global_cfg.direct_forces)
        ):
            self.norm_ffn_edge = get_normalization_layer(normalization)(
                global_cfg.hidden_size, dtype=self.backbone_dtype
            )
        else:
            self.norm_ffn_edge = nn.Identity()

        # Stochastic depth
        self.stochastic_depth_attn = (
            StochasticDepth(reg_cfg.stochastic_depth_prob)
            if reg_cfg.stochastic_depth_prob > 0.0
            else SkipStochasticDepth()
        )
        self.stochastic_depth_ffn = (
            StochasticDepth(reg_cfg.stochastic_depth_prob)
            if reg_cfg.stochastic_depth_prob > 0.0
            else SkipStochasticDepth()
        )

    def forward(
        self,
        data: GraphAttentionData,
        node_features: torch.Tensor,
        edge_features: torch.Tensor,
    ):
        # ref: swin transformer https://github.com/pytorch/vision/blob/main/torchvision/models/swin_transformer.py#L452
        # x = x + self.stochastic_depth(self.graph_attention(self.norm_attn(x)))
        # x = x + self.stochastic_depth(self.feedforward(self.norm_ffn(x)))

        # attention
        node_hidden, edge_hidden = (
            self.norm_attn_node(node_features),
            self.norm_attn_edge(edge_features),
        )
        node_hidden, edge_hidden = self.graph_attention(data, node_hidden, edge_hidden)
        node_hidden, edge_hidden = self.stochastic_depth_attn(
            node_hidden, edge_hidden, data.node_batch
        )
        node_features, edge_features = (
            node_hidden + node_features,
            edge_hidden + edge_features,
        )

        # feedforward
        node_hidden, edge_hidden = (
            self.norm_ffn_node(node_features),
            self.norm_ffn_edge(edge_features),
        )
        node_hidden, edge_hidden = self.feedforward(node_hidden, edge_hidden)
        node_hidden, edge_hidden = self.stochastic_depth_ffn(
            node_hidden, edge_hidden, data.node_batch
        )
        node_features, edge_features = (
            node_hidden + node_features,
            edge_hidden + edge_features,
        )
        return node_features, edge_features


class EfficientGraphAttention(BaseGraphNeuralNetworkLayer):
    """
    Efficient Graph Attention module.
    """

    def __init__(
        self,
        global_cfg: GlobalConfigs,
        molecular_graph_cfg: MolecularGraphConfigs,
        gnn_cfg: GraphNeuralNetworksConfigs,
        reg_cfg: RegularizationConfigs,
    ):
        super().__init__(global_cfg, molecular_graph_cfg, gnn_cfg, reg_cfg)

        self.backbone_dtype = (
            torch.float16 if global_cfg.use_fp16_backbone else torch.float32
        )

        # Convert repeating_dimensions to Python list for compile-time constants
        # This helps torch.compile recognize these as static values
        self.repeating_dimensions_list = gnn_cfg.freequency_list

        # Also store the length as a constant for looping
        self.rep_dim_len = len(self.repeating_dimensions_list)

        # Register buffer for use in non-compiled contexts
        repeating_dimensions = torch.tensor(gnn_cfg.freequency_list, dtype=torch.long)
        self.register_buffer(
            "repeating_dimensions", repeating_dimensions, persistent=False
        )

        # Pre-calculate the padding size needed for memory-efficient attention
        if gnn_cfg.use_frequency_embedding:
            # Calculate the total dimension of the expanded frequency vectors
            freq_dim = 0
            for _l, rep_count in enumerate(gnn_cfg.freequency_list):
                if rep_count > 0:
                    freq_dim += rep_count * (2 * _l + 1)

            # Calculate padding needed to make it divisible by 8
            padding_size = (8 - freq_dim % 8) % 8
            self.padding_size = padding_size
            self.register_buffer(
                "padding_size_tensor",
                torch.tensor(padding_size, dtype=torch.long),
                persistent=False,
            )

        # Store use_frequency_embedding flag
        self.use_frequency_embedding = gnn_cfg.use_frequency_embedding

        # Edge linear layer
        self.edge_attr_linear = self.get_edge_linear(gnn_cfg, global_cfg, reg_cfg)
        self.edge_attr_norm = get_normalization_layer(
            NormalizationType(reg_cfg.normalization)
        )(global_cfg.hidden_size)

        # Node hidden layer
        self.node_hidden_linear = self.get_node_linear(global_cfg, reg_cfg).to(
            self.backbone_dtype
        )

        # Edge hidden layer
        self.edge_hidden_linear = get_linear(
            in_features=global_cfg.hidden_size,
            out_features=global_cfg.hidden_size,
            activation=Activation(global_cfg.activation),
            bias=True,
            dropout=reg_cfg.mlp_dropout,
        ).to(self.backbone_dtype)

        # message normalization
        self.message_norm = get_normalization_layer(
            NormalizationType(reg_cfg.normalization)
        )(global_cfg.hidden_size, dtype=self.backbone_dtype)

        # message linear
        self.use_message_gate = gnn_cfg.use_message_gate
        if self.use_message_gate:
            self.gate_linear = get_linear(
                in_features=global_cfg.hidden_size * 2,
                out_features=global_cfg.hidden_size,
                activation=None,
                bias=True,
            ).to(self.backbone_dtype)
            self.candidate_linear = get_linear(
                in_features=global_cfg.hidden_size * 2,
                out_features=global_cfg.hidden_size,
                activation=None,
                bias=True,
            ).to(self.backbone_dtype)
        else:
            self.message_linear = get_linear(
                in_features=global_cfg.hidden_size * 3,
                out_features=global_cfg.hidden_size,
                activation=Activation(global_cfg.activation),
                bias=True,
            ).to(self.backbone_dtype)

        # Multi-head attention
        # self.multi_head_attention = nn.MultiheadAttention(
        #     embed_dim=global_cfg.hidden_size,
        #     num_heads=gnn_cfg.atten_num_heads,
        #     dropout=reg_cfg.atten_dropout,
        #     bias=True,
        #     batch_first=True,
        #     dtype=self.backbone_dtype,
        # )
        self.attn_in_proj_q = nn.Linear(
            global_cfg.hidden_size,
            global_cfg.hidden_size,
            bias=True,
        ).to(self.backbone_dtype)
        self.attn_in_proj_k = nn.Linear(
            global_cfg.hidden_size,
            global_cfg.hidden_size,
            bias=True,
        ).to(self.backbone_dtype)
        self.attn_in_proj_v = nn.Linear(
            global_cfg.hidden_size,
            global_cfg.hidden_size,
            bias=True,
        ).to(self.backbone_dtype)
        self.attn_out_proj = nn.Linear(
            global_cfg.hidden_size,
            global_cfg.hidden_size,
            bias=True,
        ).to(self.backbone_dtype)
        self.attn_num_heads = gnn_cfg.atten_num_heads
        self.attn_dropout = reg_cfg.atten_dropout

        # scalar for attention bias
        self.use_angle_embedding = gnn_cfg.use_angle_embedding
        if self.use_angle_embedding == "scalar":
            self.attn_scalar = nn.Parameter(
                torch.ones(gnn_cfg.atten_num_heads, dtype=self.backbone_dtype),
                requires_grad=True,
            )
        elif self.use_angle_embedding == "bias":
            self.attn_distance_embedding = get_linear(
                in_features=gnn_cfg.edge_distance_expansion_size,
                out_features=gnn_cfg.angle_embedding_size,
                activation=Activation(global_cfg.activation),
                bias=True,
                dropout=reg_cfg.mlp_dropout,
            )
            self.attn_angle_embedding = get_linear(
                in_features=gnn_cfg.angle_expansion_size + 1,
                out_features=gnn_cfg.angle_embedding_size,
                activation=Activation(global_cfg.activation),
                bias=True,
                dropout=reg_cfg.mlp_dropout,
            )
            self.attn_bias_projection = get_linear(
                in_features=gnn_cfg.angle_embedding_size**3,
                out_features=gnn_cfg.atten_num_heads,
                activation=None,
                bias=True,
                dropout=reg_cfg.mlp_dropout,
            )
        elif self.use_angle_embedding == "none":
            self.attn_scalar = torch.tensor(1.0, dtype=self.backbone_dtype)
        else:
            raise ValueError(
                f"Invalid use_angle_embedding {gnn_cfg.use_angle_embedding}. Must be one of ['scalar', 'bias', 'none']"
            )

        # Graph attention for aggregation
        # ref: "How Attentive are Graph Attention Networks?" <https://arxiv.org/abs/2105.14491>
        self.use_graph_attention = gnn_cfg.use_graph_attention
        if self.use_graph_attention:
            self.attn_weight = nn.Parameter(
                torch.empty(
                    1,
                    1,
                    gnn_cfg.atten_num_heads,
                    global_cfg.hidden_size // gnn_cfg.atten_num_heads,
                    dtype=self.backbone_dtype,
                ),
                requires_grad=True,
            )
            # glorot initialization
            stdv = math.sqrt(
                6.0 / (self.attn_weight.shape[-2] + self.attn_weight.shape[-1])
            )
            self.attn_weight.data.uniform_(-stdv, stdv)

    def forward(
        self,
        data: GraphAttentionData,
        node_features: torch.Tensor,
        edge_features: torch.Tensor,
    ):
        # Get edge attributes
        edge_attr = self.get_edge_features(data)

        edge_attr = self.edge_attr_linear(edge_attr)
        edge_attr = self.edge_attr_norm(edge_attr)
        edge_attr = edge_attr.to(self.backbone_dtype)

        # Get node features
        node_features = self.get_node_features(node_features, data.neighbor_list)
        node_hidden = self.node_hidden_linear(node_features)

        # Get edge faetures
        edge_hidden = self.edge_hidden_linear(edge_features)

        # Concatenate edge and node features (num_nodes, num_neighbors, hidden_size)
        if self.use_message_gate:
            message = torch.cat([edge_attr, node_hidden], dim=-1)
            update_gate = torch.sigmoid(self.gate_linear(message))
            candidate = torch.tanh(self.candidate_linear(message))
            message = update_gate * candidate + (1 - update_gate) * edge_hidden
        else:
            message = self.message_linear(
                torch.cat([edge_attr, edge_hidden, node_hidden], dim=-1)
            )
        message = self.message_norm(message)

        # Multi-head self-attention
        if self.use_angle_embedding == "bias":
            attn_mask = data.attn_mask + self.get_attn_bias(
                data.angle_embedding, data.edge_distance_expansion
            )
        elif self.use_angle_embedding == "scalar" and data.angle_embedding is not None:
            angle_embedding = data.angle_embedding.reshape(
                -1,
                self.attn_scalar.shape[0],
                data.angle_embedding.shape[-2],
                data.angle_embedding.shape[-1],
            ) * self.attn_scalar.unsqueeze(0).unsqueeze(-1).unsqueeze(-1)
            angle_embedding = angle_embedding.reshape(
                -1, data.angle_embedding.shape[-2], data.angle_embedding.shape[-1]
            )
            attn_mask = data.attn_mask + angle_embedding
        else:
            attn_mask = data.attn_mask
        # edge_output = self.multi_head_attention(
        #     query=message,
        #     key=message,
        #     value=message,
        #     # key_padding_mask=~data.neighbor_mask,
        #     attn_mask=attn_mask,
        #     need_weights=False,
        # )[0]
        edge_output = self.multi_head_self_attention(
            input=message,
            attn_mask=attn_mask,
            frequency_vectors=data.frequency_vectors,
        )

        # Aggregation
        if self.use_graph_attention:
            node_output, edge_output = self.graph_attention_aggregate(
                edge_output, data.neighbor_mask
            )
        else:
            node_output = self.aggregate(
                edge_output, data.neighbor_mask
            )  ##### TODO: use masked with envolope

        return node_output, edge_output

    def multi_head_self_attention(self, input, attn_mask, frequency_vectors=None):
        # input (num_nodes, num_neighbors, hidden_size)
        # attn_mask (num_nodes * num_heads, num_neighbors, num_neighbors)
        # frequency_vectors: (num_nodes, num_neighbors, sum_{l=0..lmax} rep_l * (2l+1))
        num_nodes, num_neighbors, hidden_dim = input.shape
        head_dim = hidden_dim // self.attn_num_heads
        q = (
            self.attn_in_proj_q(input)
            .reshape(num_nodes, num_neighbors, self.attn_num_heads, head_dim)
            .permute(0, 2, 1, 3)
        )
        k = (
            self.attn_in_proj_k(input)
            .reshape(num_nodes, num_neighbors, self.attn_num_heads, head_dim)
            .permute(0, 2, 1, 3)
        )
        v = (
            self.attn_in_proj_v(input)
            .reshape(num_nodes, num_neighbors, self.attn_num_heads, head_dim)
            .permute(0, 2, 1, 3)
        )
        # q,k,v (num_nodes, num_heads, num_neighbors, head_dim)

        # Apply frequency embedding if enabled
        if self.use_frequency_embedding and frequency_vectors is not None:
            # Add head dimension to frequency vectors
            # (num_nodes, num_neighbors, sum_{l=0..lmax} rep_l * (2l+1)) ->
            # (num_nodes, 1, num_neighbors, sum_{l=0..lmax} rep_l * (2l+1))
            freq_vecs = frequency_vectors.unsqueeze(1)

            # Create expanded q and k by repeating sections according to repeating_dimensions
            # For each l-value, we repeat the corresponding section 2*l+1 times

            # Lists to collect expanded sections
            q_expanded_sections = []
            k_expanded_sections = []

            # Current position in the head_dim
            curr_pos = 0

            # For each l value - use Python constant for loop range
            for _l in range(self.rep_dim_len):
                # Get repeat count from the Python list - not a tensor
                rep_count = self.repeating_dimensions_list[_l]

                # Skip zero repeats - this is now a static check during compilation
                if rep_count == 0:
                    continue

                # Skip if we've reached the end of the head_dim
                if curr_pos >= head_dim:
                    break

                # Calculate repetition factor for this l: 2*l+1
                sh_dim = 2 * _l + 1

                # End position for this segment
                end_pos = min(curr_pos + rep_count, head_dim)

                # Get the corresponding section from q and k
                q_section = q[
                    ..., curr_pos:end_pos
                ]  # (num_nodes, num_heads, num_neighbors, rep_count)
                k_section = k[
                    ..., curr_pos:end_pos
                ]  # (num_nodes, num_heads, num_neighbors, rep_count)

                # Reshape to prepare for repeating each dimension
                # (num_nodes, num_heads, num_neighbors, rep_count) -> (num_nodes, num_heads, num_neighbors, rep_count, 1)
                q_section = q_section.unsqueeze(-1)
                k_section = k_section.unsqueeze(-1)

                # Repeat each dimension 2*l+1 times
                # (num_nodes, num_heads, num_neighbors, rep_count, 1) -> (num_nodes, num_heads, num_neighbors, rep_count, 2*l+1)
                q_expanded = q_section.expand(-1, -1, -1, -1, sh_dim)
                k_expanded = k_section.expand(-1, -1, -1, -1, sh_dim)

                # Reshape to flatten the last two dimensions
                # (num_nodes, num_heads, num_neighbors, rep_count, 2*l+1) -> (num_nodes, num_heads, num_neighbors, rep_count*(2*l+1))
                q_expanded = q_expanded.reshape(
                    num_nodes, self.attn_num_heads, num_neighbors, -1
                )
                k_expanded = k_expanded.reshape(
                    num_nodes, self.attn_num_heads, num_neighbors, -1
                )

                # Add to our collection
                q_expanded_sections.append(q_expanded)
                k_expanded_sections.append(k_expanded)

                # Move to the next position
                curr_pos = end_pos

            # Only process if we have expanded sections
            if q_expanded_sections:
                # Concatenate the expanded sections
                # [(num_nodes, num_heads, num_neighbors, rep_0*(2*0+1)), (num_nodes, num_heads, num_neighbors, rep_1*(2*1+1)), ...]
                # -> (num_nodes, num_heads, num_neighbors, sum_l rep_l*(2*l+1))
                q = torch.cat(q_expanded_sections, dim=-1)
                k = torch.cat(k_expanded_sections, dim=-1)

                # Apply frequency vectors
                q = q * freq_vecs
                k = k * freq_vecs

                # Pad q and k to make their last dimension divisible by 8 for memory-efficient attention
                # Using Python constant instead of tensor
                if self.padding_size > 0:
                    # Pad the last dimension with zeros
                    q = F.pad(q, (0, self.padding_size))
                    k = F.pad(k, (0, self.padding_size))
                    # Also pad the frequency vectors to match
                    freq_vecs = F.pad(freq_vecs, (0, self.padding_size))

                # Scale appropriately
                # constant = math.sqrt(q.shape[-1] / head_dim)
                # q = q * constant
                # k = k * constant

        # View attention mask
        attn_mask = attn_mask.view(
            num_nodes, self.attn_num_heads, num_neighbors, num_neighbors
        )

        dropout_p = self.attn_dropout if self.training else 0.0

        attn_output = F.scaled_dot_product_attention(
            query=q,
            key=k,
            value=v,
            attn_mask=attn_mask,
            dropout_p=dropout_p,
        )

        attn_output = attn_output.permute(0, 2, 1, 3).reshape(
            num_nodes, num_neighbors, hidden_dim
        )

        attn_output = self.attn_out_proj(attn_output)

        return attn_output

    def get_attn_bias(self, angle_embedding, edge_distance_expansion):
        # legendre_embedding (num_nodes, num_neighbors, num_neighbors, angle_embedding_size)
        legendre_embedding = self.attn_angle_embedding(angle_embedding)
        # edge_distance_expansion (num_nodes, num_neighbors, angle_embedding_size)
        edge_distance_embedding = self.attn_distance_embedding(edge_distance_expansion)
        # edge_dist_outer (num_nodes, num_neighbors, num_neighbors, angle_embedding_size)
        edge_dist_outer = edge_distance_embedding.unsqueeze(
            2
        ) * edge_distance_embedding.unsqueeze(1)
        # edge_dist_outer (num_nodes, num_neighbors, num_neighbors, angle_embedding_size, angle_embedding_size)
        edge_dist_outer = edge_dist_outer.unsqueeze(-1) * edge_dist_outer.unsqueeze(-2)
        # angle_embedding (num_nodes, num_neighbors, num_neighbors, angle_embedding_size, angle_embedding_size, angle_embedding_size)
        angle_embedding = legendre_embedding.unsqueeze(-1).unsqueeze(
            -1
        ) * edge_dist_outer.unsqueeze(-3)
        # angle_embedding (num_nodes, num_neighbors, num_neighbors, angle_embedding_size ** 3)
        angle_embedding = angle_embedding.view(
            angle_embedding.shape[0],
            angle_embedding.shape[1],
            angle_embedding.shape[2],
            -1,
        )
        # attn_bias (num_nodes, num_neighbors, num_neighbors, num_heads)
        attn_bias = self.attn_bias_projection(angle_embedding)
        # attn_bias (num_nodes * num_heads, max_neighbors, max_neighbors)
        attn_bias = attn_bias.permute(0, 3, 1, 2).reshape(
            -1, attn_bias.shape[1], attn_bias.shape[2]
        )
        return attn_bias

    def graph_attention_aggregate(self, edge_output, neighbor_mask):
        # Graph attention for aggregation
        # ref: "How Attentive are Graph Attention Networks?" <https://arxiv.org/abs/2105.14491>

        num_nodes, num_neighbors, _ = edge_output.shape
        _, _, num_heads, head_dim = self.attn_weight.shape

        edge_output = edge_output.view(num_nodes, num_neighbors, num_heads, head_dim)
        # alpha (num_nodes, num_neighbors, num_heads)
        alpha = (edge_output * self.attn_weight).sum(-1)
        alpha = F.leaky_relu(alpha, negative_slope=0.2)
        alpha = alpha.masked_fill(neighbor_mask.unsqueeze(-1) == 0, float("-inf"))
        alpha = F.softmax(alpha, dim=1)
        node_output = (alpha.unsqueeze(-1) * edge_output).sum(1)
        node_output = node_output.view(num_nodes, -1)
        edge_output = edge_output.view(num_nodes, num_neighbors, -1)
        return node_output, edge_output


class FeedForwardNetwork(nn.Module):
    """
    Feed Forward Network module.
    """

    def __init__(
        self,
        global_cfg: GlobalConfigs,
        gnn_cfg: GraphNeuralNetworksConfigs,
        reg_cfg: RegularizationConfigs,
        is_last: bool = False,
    ):
        super().__init__()
        self.backbone_dtype = (
            torch.float16 if global_cfg.use_fp16_backbone else torch.float32
        )
        self.mlp_node = get_feedforward(
            hidden_dim=global_cfg.hidden_size,
            activation=Activation(global_cfg.activation),
            hidden_layer_multiplier=gnn_cfg.ffn_hidden_layer_multiplier,
            bias=True,
            dropout=reg_cfg.node_ffn_dropout,
        ).to(self.backbone_dtype)
        if not (
            (is_last) and (global_cfg.regress_forces and not global_cfg.direct_forces)
        ):
            self.mlp_edge = get_feedforward(
                hidden_dim=global_cfg.hidden_size,
                activation=Activation(global_cfg.activation),
                hidden_layer_multiplier=gnn_cfg.ffn_hidden_layer_multiplier,
                bias=True,
                dropout=reg_cfg.edge_ffn_dropout,
            ).to(self.backbone_dtype)
        else:
            self.mlp_edge = nn.Identity()

    def forward(self, node_features: torch.Tensor, edge_features: torch.Tensor):
        return self.mlp_node(node_features), self.mlp_edge(edge_features)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

Original source: https://github.com/janosh/matbench-discovery/blob/main/matbench_discovery/energy.py

MIT License

Copyright (c) 2022 Janosh Riebesell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

The software is provided "as is", without warranty of any kind, express or
implied, including but not limited to the warranties of merchantability,
fitness for a particular purpose and noninfringement. In no event shall the
authors or copyright holders be liable for any claim, damages or other
liability, whether in an action of contract, tort or otherwise, arising from,
out of or in connection with the software or the use or other dealings in the
software.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from monty.dev import requires

try:
    from pymatgen.analysis.phase_diagram import Entry
    from pymatgen.core import Composition, Structure

    pmg_installed = True
except ImportError:
    Entry, Composition, Structure = None, None, None
    pmg_installed = False


if TYPE_CHECKING:
    from pymatgen.util.typing import EntryLike


@requires(pmg_installed, message="Requires `pymatgen` to be installed")
def calc_energy_from_e_refs(
    struct_or_entry: EntryLike | Structure | Composition | str,
    ref_energies: dict[str, float],
    total_energy: float | None = None,
) -> float:
    """Calculate energy per atom relative to reference states (e.g., for formation or
    cohesive energy calculations).

    Args:
        struct_or_entry (EntryLike | Structure | Composition | str): Either:
            - A pymatgen Entry (PDEntry, ComputedEntry, etc.) or entry dict containing
              'energy' and 'composition' keys
            - A Structure or Composition object or formula string (must also provide
              total_energy)
        ref_energies (dict[str, float]): Dictionary of reference energies per atom.
            For formation energy: elemental reference energies (e.g.
            mp_elemental_ref_energies).
            For cohesive energy: isolated atom reference energies
        total_energy (float | None): Total energy of the structure/composition. Required
            if struct_or_entry is not an Entry or entry dict. Ignored otherwise.

    Returns:
        float: Energy per atom relative to references (e.g., formation or cohesive
        energy) in the same units as input energies.

    Raises:
        TypeError: If input types are invalid
        ValueError: If missing reference energies for some elements
    """
    if isinstance(struct_or_entry, dict):  # entry dict case
        energy = struct_or_entry["energy"]
        comp = Composition(struct_or_entry["composition"])
    # Entry/ComputedEntry/ComputedStructureEntry instance case
    elif isinstance(struct_or_entry, Entry):
        energy = struct_or_entry.energy
        comp = struct_or_entry.composition
    else:  # Structure/Composition/formula case
        if total_energy is None:
            raise ValueError("total_energy can't be None when 1st arg is not an Entry")
        energy = total_energy

        if isinstance(struct_or_entry, str):
            comp = Composition(struct_or_entry)
        elif isinstance(struct_or_entry, Structure):
            comp = struct_or_entry.composition
        elif isinstance(struct_or_entry, Composition):
            comp = struct_or_entry
        else:
            cls_name = type(struct_or_entry).__name__
            raise TypeError(
                "Expected Entry, Structure, Composition or formula string, "
                f"got {cls_name}"
            )

    # Check that we have all needed reference energies
    if missing_refs := set(map(str, comp)) - set(ref_energies):
        raise ValueError(f"Missing reference energies for elements: {missing_refs}")

    # Calculate reference energy
    e_ref = sum(ref_energies[str(el)] * amt for el, amt in comp.items())

    return (energy - e_ref) / comp.num_atoms



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import copy
import logging
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

import torch
from torch import nn

from fairchem.core.common.registry import registry
from fairchem.core.common.utils import (
    load_model_and_weights_from_checkpoint,
)

if TYPE_CHECKING:
    from fairchem.core.datasets.atomic_data import AtomicData


class HeadInterface(metaclass=ABCMeta):
    @property
    def use_amp(self):
        return False

    @abstractmethod
    def forward(
        self, data: AtomicData, emb: dict[str, torch.Tensor]
    ) -> dict[str, torch.Tensor]:
        """Head forward.

        Arguments
        ---------
        data: AtomicData
            Atomic systems as input
        emb: dict[str->torch.Tensor]
            Embeddings of the input as generated by the backbone

        Returns
        -------
        outputs: dict[str->torch.Tensor]
            Return one or more targets generated by this head
        """
        return


class BackboneInterface(metaclass=ABCMeta):
    @abstractmethod
    def forward(self, data: AtomicData) -> dict[str, torch.Tensor]:
        """Backbone forward.

        Arguments
        ---------
        data: AtomicData
            Atomic systems as input

        Returns
        -------
        embedding: dict[str->torch.Tensor]
            Return backbone embeddings for the given input
        """
        return


@registry.register_model("hydra")
class HydraModel(nn.Module):
    def __init__(
        self,
        backbone: dict | None = None,
        heads: dict | None = None,
        finetune_config: dict | None = None,
        otf_graph: bool = True,
        pass_through_head_outputs: bool = False,
        freeze_backbone: bool = False,
    ):
        super().__init__()
        self.device = None
        self.otf_graph = otf_graph
        # This is required for hydras with models that have multiple outputs per head, since we will deprecate
        # the old config system at some point, this will prevent the need to make major modifications to the trainer
        # because they all expect the name of the outputs directly instead of the head_name.property_name
        self.pass_through_head_outputs = pass_through_head_outputs

        # if finetune_config is provided, then attempt to load the model from the given finetune checkpoint
        starting_model = None
        if finetune_config is not None:
            # Make it hard to sneak more fields into finetuneconfig
            assert (
                len(set(finetune_config.keys()) - {"starting_checkpoint", "override"})
                == 0
            )
            starting_model: HydraModel = load_model_and_weights_from_checkpoint(
                finetune_config["starting_checkpoint"]
            )
            logging.info(
                f"Found and loaded fine-tuning checkpoint: {finetune_config['starting_checkpoint']} (Note we are NOT loading the training state from this checkpoint, only parts of the model and weights)"
            )
            assert isinstance(
                starting_model, HydraModel
            ), "Can only finetune starting from other hydra models!"
            # TODO this is a bit hacky to overrride attrs in the backbone
            if "override" in finetune_config:
                for key, value in finetune_config["override"].items():
                    setattr(starting_model.backbone, key, value)

        if backbone is not None:
            backbone = copy.deepcopy(backbone)
            backbone_model_name = backbone.pop("model")
            self.backbone: BackboneInterface = registry.get_model_class(
                backbone_model_name
            )(
                **backbone,
            )
        elif starting_model is not None:
            self.backbone = starting_model.backbone
            logging.info(
                f"User did not specify a backbone, using the backbone from the starting checkpoint {self.backbone}"
            )
        else:
            raise RuntimeError(
                "Backbone not specified and not found in the starting checkpoint"
            )

        if freeze_backbone:
            for param in self.backbone.parameters():
                param.requires_grad = False

        if heads is not None:
            heads = copy.deepcopy(heads)
            # Iterate through outputs_cfg and create heads
            self.output_heads: dict[str, HeadInterface] = {}

            head_names_sorted = sorted(heads.keys())
            assert len(set(head_names_sorted)) == len(
                head_names_sorted
            ), "Head names must be unique!"
            for head_name in head_names_sorted:
                head_config = heads[head_name]
                if "module" not in head_config:
                    raise ValueError(
                        f"{head_name} head does not specify module to use for the head"
                    )

                module_name = head_config.pop("module")
                self.output_heads[head_name] = registry.get_model_class(module_name)(
                    self.backbone,
                    **head_config,
                )

            self.output_heads = torch.nn.ModuleDict(self.output_heads)
        elif starting_model is not None:
            self.output_heads = starting_model.output_heads
            logging.info(
                f"User did not specify heads, using the output heads from the starting checkpoint {self.output_heads}"
            )
        else:
            raise RuntimeError(
                "Heads not specified and not found in the starting checkpoint"
            )

    def forward(self, data: AtomicData):
        # lazily get device from input to use with amp, at least one input must be a tensor to figure out it's device
        if not self.device:
            device_from_tensors = {
                x.device.type for x in data.values() if isinstance(x, torch.Tensor)
            }
            assert (
                len(device_from_tensors) == 1
            ), f"all inputs must be on the same device, found the following devices {device_from_tensors}"
            self.device = device_from_tensors.pop()

        emb = self.backbone(data)
        # Predict all output properties for all structures in the batch for now.
        out = {}
        for k in self.output_heads:
            with torch.autocast(
                device_type=self.device, enabled=self.output_heads[k].use_amp
            ):
                if self.pass_through_head_outputs:
                    out.update(self.output_heads[k](data, emb))
                else:
                    out[k] = self.output_heads[k](data, emb)

        return out


class HydraModelV2(nn.Module):
    def __init__(
        self,
        backbone: BackboneInterface,
        heads: dict[str, HeadInterface],
        freeze_backbone: bool = False,
    ):
        super().__init__()
        self.backbone = backbone
        self.output_heads = torch.nn.ModuleDict(heads)
        self.device = None
        if freeze_backbone:
            for param in self.backbone.parameters():
                param.requires_grad = False

    def forward(self, data):
        # lazily get device from input to use with amp, at least one input must be a tensor to figure out it's device
        if not self.device:
            device_from_tensors = {
                x.device.type for x in data.values() if isinstance(x, torch.Tensor)
            }
            assert (
                len(device_from_tensors) == 1
            ), f"all inputs must be on the same device, found the following devices {device_from_tensors}"
            self.device = device_from_tensors.pop()

        emb = self.backbone(data)
        # Predict all output properties for all structures in the batch for now.
        out = {}
        for k in self.output_heads:
            with torch.autocast(
                device_type=self.device, enabled=self.output_heads[k].use_amp
            ):
                out[k] = self.output_heads[k](data, emb)
        return out



from __future__ import annotations

from enum import Enum
from typing import Optional

import torch
import torch.nn as nn

## The following part is from xformers, copied here in case it's deprecated
## Ref: https://github.com/facebookresearch/xformers/blob/main/xformers/components/activations.py


class Activation(str, Enum):
    SquaredReLU = "squared_relu"
    GeLU = "gelu"
    LeakyReLU = "leaky_relu"
    ReLU = "relu"
    SmeLU = "smelu"
    StarReLU = "star_relu"


# For unit testing / parity comparisons, probably not the fastest way
class SquaredReLU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_ = torch.nn.functional.relu(x)
        return x_ * x_


class StarReLU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_ = torch.nn.functional.relu(x)
        return 0.8944 * x_ * x_ - 0.4472


class SmeLU(nn.Module):
    def __init__(self, beta: float = 2.0) -> None:
        super().__init__()
        self.beta = beta

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        relu = torch.where(
            x >= self.beta,
            x,
            torch.tensor([0.0], device=x.device, dtype=x.dtype),
        )
        return torch.where(
            torch.abs(x) <= self.beta,
            ((x + self.beta) ** 2).type_as(x) / (4.0 * self.beta),
            relu,
        )


def build_activation(activation: Optional[Activation]):
    if not activation:
        return nn.Identity()

    return {
        Activation.ReLU: nn.ReLU,
        Activation.GeLU: nn.GELU,
        Activation.LeakyReLU: nn.LeakyReLU,
        Activation.SquaredReLU: SquaredReLU,
        Activation.StarReLU: StarReLU,
        Activation.SmeLU: SmeLU,
    }[activation]()


## End of xformers part


def get_linear(
    in_features: int,
    out_features: int,
    bias: bool = False,
    activation: Activation | None = None,
    dropout: float = 0.0,
):
    """
    Build a linear layer with optional activation and dropout.
    """
    layers: list[nn.Module] = [
        nn.Linear(in_features=in_features, out_features=out_features, bias=bias)
    ]
    if activation:
        layers.append(build_activation(activation))
    if dropout > 0.0:
        layers.append(nn.Dropout(dropout))
    return nn.Sequential(*layers)


def get_feedforward(
    hidden_dim: int,
    activation: Activation | None,
    hidden_layer_multiplier: int,
    bias: bool = False,
    dropout: float = 0.0,
):
    """
    Build a feedforward layer with optional activation function.
    """
    if hidden_layer_multiplier == 0:
        return get_linear(
            in_features=hidden_dim,
            out_features=hidden_dim,
            bias=bias,
            activation=None,
            dropout=dropout,
        )
    return nn.Sequential(
        get_linear(
            in_features=hidden_dim,
            out_features=hidden_dim * hidden_layer_multiplier,
            bias=bias,
            activation=activation,
            dropout=dropout,
        ),
        get_linear(
            in_features=hidden_dim * hidden_layer_multiplier,
            out_features=hidden_dim,
            bias=bias,
            activation=None,
            dropout=dropout,
        ),
    )


def no_weight_decay(model):
    # no weight decay on layer norms and embeddings
    # ref: https://discuss.pytorch.org/t/weight-decay-in-the-optimizers-is-a-bad-idea-especially-with-batchnorm/16994
    no_wd_list = []
    named_parameters_list = [name for name, _ in model.named_parameters()]
    for module_name, module in model.named_modules():
        if isinstance(module, (nn.Linear, nn.Embedding, nn.LayerNorm, nn.RMSNorm)):
            for parameter_name, _ in module.named_parameters():
                if isinstance(module, torch.nn.Linear) and "weight" in parameter_name:
                    continue
                global_parameter_name = module_name + "." + parameter_name
                assert global_parameter_name in named_parameters_list
                no_wd_list.append(global_parameter_name)
    return set(no_wd_list)


def init_linear_weights(module, gain=1.0):
    if isinstance(module, nn.Linear):
        nn.init.xavier_uniform_(module.weight, gain=gain)
        if module.bias is not None:
            module.bias.data.zero_()


## The following part is modified from xformers
## Ref: https://github.com/facebookresearch/xformers/blob/main/xformers/components/residual.py


class NormalizationType(str, Enum):
    LayerNorm = "layernorm"
    Skip = "skip"
    RMSNorm = "rmsnorm"


class Skip(nn.Module):
    def __init__(self, *_, **__) -> None:
        super().__init__()

    def forward(self, x, **_):
        return x


def get_normalization_layer(normalization_type: NormalizationType):
    return {
        NormalizationType.Skip: Skip,
        NormalizationType.LayerNorm: nn.LayerNorm,
        NormalizationType.RMSNorm: nn.RMSNorm,
    }[normalization_type]



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging


def get_flops_profile(model, input_data, verbose: bool = False):
    try:
        from flops_profiler.profiler import FlopsProfiler
    except Exception as e:
        logging.error(
            "To use this feature you need to install the flops profiler, pip install pip install flops-profiler"
        )
        raise e
    prof = FlopsProfiler(model)
    prof.start_profile()
    model(input_data)
    prof.stop_profile()
    flops = prof.get_total_flops()
    if verbose:
        prof.print_model_profile(profile_step=1)
    return flops



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import torch
from ase import Atoms

if TYPE_CHECKING:
    from fairchem.core.datasets.atomic_data import AtomicData
from fairchem.core.models.utils.irreps import cg_change_mat, irreps_sum


def _get_molecule_cell(data_object: AtomicData):
    # create an Atoms object and center molecule in cell
    mol = Atoms(
        numbers=data_object.atomic_numbers,
        positions=data_object.pos,
    )
    # largest radius cutoff is ~12A include that and a safety factor of 10
    mol.center(vacuum=(12.0 * 10.0))
    mol.pbc = [True, True, True]

    positions = np.array(mol.get_positions(), copy=True)
    # pbc = np.array(mol.pbc, copy=True)
    cell = np.array(mol.get_cell(complete=True), copy=True)

    atomic_numbers = torch.Tensor(mol.get_atomic_numbers())
    positions = torch.from_numpy(positions).float()
    cell = torch.from_numpy(cell).view(1, 3, 3).float()
    natoms = positions.shape[0]
    assert data_object.natoms == natoms

    return atomic_numbers, positions, cell


def common_transform(data_object: AtomicData, config) -> AtomicData:
    data_object.dataset = config["dataset_name"]

    if not hasattr(data_object, "charge"):
        data_object.charge = 0
    if not hasattr(data_object, "spin"):
        data_object.spin = 0
    ensure_tensor(data_object, "energy")
    return data_object


def ensure_tensor(data_object, keys):
    # ensure dataset_energy is a tensor
    if isinstance(keys, str):
        keys = [keys]
    for key in keys:
        if hasattr(data_object, key):
            if not torch.is_tensor(getattr(data_object, key)):
                setattr(
                    data_object,
                    key,
                    torch.tensor(getattr(data_object, key), dtype=torch.float),
                )
            setattr(data_object, key, getattr(data_object, key).view(-1).float())
    return data_object


def ani1x_transform(data_object: AtomicData, config) -> AtomicData:
    # make periodic with molecule centered in large cell
    atomic_numbers, positions, cell = _get_molecule_cell(data_object)
    data_object.atomic_numbers = atomic_numbers
    data_object.pos = positions
    data_object.cell = cell

    # add fixed
    data_object.fixed = torch.zeros(data_object.natoms, dtype=torch.float)

    # common transforms
    data_object = common_transform(data_object, config)

    # ensure ani1x_energy is a tensor
    return data_object


def trans1x_transform(data_object: AtomicData, config) -> AtomicData:
    # make periodic with molecule centered in large cell
    atomic_numbers, positions, cell = _get_molecule_cell(data_object)
    data_object.atomic_numbers = atomic_numbers
    data_object.pos = positions
    data_object.cell = cell

    # add fixed and cell
    data_object.fixed = torch.zeros(data_object.natoms, dtype=torch.float)

    # common transforms
    data_object = common_transform(data_object, config)

    # ensure trans1x_energy is a tensor

    return data_object


def spice_transform(data_object: AtomicData, config) -> AtomicData:
    # make periodic with molecule centered in large cell
    atomic_numbers, positions, cell = _get_molecule_cell(data_object)
    data_object.atomic_numbers = atomic_numbers
    data_object.pos = positions
    data_object.cell = cell

    # add fixed
    data_object.fixed = torch.zeros(data_object.natoms, dtype=torch.float)

    # common transforms
    data_object = common_transform(data_object, config)
    # this is necessary for SPICE maceoff split to work with GemNet-OC
    data_object.tags = torch.full(data_object.tags.shape, 2, dtype=torch.long)

    # ensure spice_energy is a tensor

    return data_object


def qmof_transform(data_object: AtomicData, config) -> AtomicData:
    # add fixed and cell
    data_object.fixed = torch.zeros(data_object.natoms, dtype=torch.float)

    # common transforms
    data_object = common_transform(data_object, config)

    return data_object


def qm9_transform(data_object: AtomicData, config) -> AtomicData:
    # make periodic with molecule centered in large cell
    atomic_numbers, positions, cell = _get_molecule_cell(data_object)
    data_object.atomic_numbers = atomic_numbers
    data_object.pos = positions
    data_object.cell = cell

    # add fixed
    data_object.fixed = torch.zeros(data_object.natoms, dtype=torch.float)

    # common transforms
    data_object = common_transform(data_object, config)

    return data_object


def omol_transform(data_object: AtomicData, config) -> AtomicData:
    # make periodic with molecule centered in large cell
    atomic_numbers, positions, cell = _get_molecule_cell(data_object)
    data_object.atomic_numbers = atomic_numbers
    data_object.pos = positions
    data_object.cell = cell
    assert hasattr(
        data_object, "charge"
    ), "no charge in omol dataset set a2g_args: {r_energy: True, r_forces: True, r_data_keys: ['spin', 'charge']}"
    assert hasattr(
        data_object, "spin"
    ), "no spin in omol dataset set a2g_args: {r_energy: True, r_forces: True, r_data_keys: ['spin', 'charge']}"

    # add fixed
    data_object.fixed = torch.zeros(data_object.natoms, dtype=torch.float)

    return common_transform(data_object, config)


def stress_reshape_transform(data_object: AtomicData, config) -> AtomicData:
    for k in data_object.keys():  # noqa: SIM118
        if "stress" in k and ("iso" not in k and "aniso" not in k):
            data_object[k] = data_object[k].reshape(1, 9)
    return data_object


def asedb_transform(data_object: AtomicData, config) -> AtomicData:
    data_object.dataset = config["dataset_name"]
    data_object.sid = str(
        data_object.sid.item() if torch.is_tensor(data_object) else data_object.sid
    )
    return data_object


class DataTransforms:
    def __init__(self, config) -> None:
        self.config = config

    def __call__(self, data_object):
        if not self.config:
            return data_object

        for transform_fn in self.config:
            # TODO: Normalization information used in the trainers. Ignore here for now
            # TODO: if we dont use them here, these should not be defined as "transforms" in the config
            # TODO: add them as another entry under dataset, maybe "standardize"?
            if transform_fn in ("normalizer", "element_references"):
                continue

            data_object = eval(transform_fn)(data_object, self.config[transform_fn])

        return data_object


def decompose_tensor(data_object, config) -> AtomicData:
    tensor_key = config["tensor"]
    rank = config["rank"]

    if tensor_key not in data_object:
        return data_object

    if rank != 2:
        raise NotImplementedError

    tensor_decomposition = torch.einsum(
        "ab, cb->ca",
        cg_change_mat(rank),
        data_object[tensor_key].reshape(1, irreps_sum(rank)),
    )

    for decomposition_key in config["decomposition"]:
        irrep_dim = config["decomposition"][decomposition_key]["irrep_dim"]
        data_object[decomposition_key] = tensor_decomposition[
            :,
            max(0, irreps_sum(irrep_dim - 1)) : irreps_sum(irrep_dim),
        ]

    return data_object



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

1. Normalize features of shape (N, sphere_basis, C),
with sphere_basis = (lmax + 1) ** 2.

2. The difference from `layer_norm.py` is that all type-L vectors have
the same number of channels and input features are of shape (N, sphere_basis, C).
"""

from __future__ import annotations

import math
from typing import Literal

import torch
import torch.nn as nn


def get_normalization_layer(
    norm_type: Literal["layer_norm", "layer_norm_sh", "rms_norm_sh"],
    lmax: int,
    num_channels: int,
    eps: float = 1e-5,
    affine: bool = True,
    normalization: str = "component",
):
    assert norm_type in ["layer_norm", "layer_norm_sh", "rms_norm_sh"]
    if norm_type == "layer_norm":
        norm_class = EquivariantLayerNormArray
    elif norm_type == "layer_norm_sh":
        norm_class = EquivariantLayerNormArraySphericalHarmonics
    elif norm_type == "rms_norm_sh":
        norm_class = EquivariantRMSNormArraySphericalHarmonicsV2
    else:
        raise ValueError
    return norm_class(lmax, num_channels, eps, affine, normalization)


def get_l_to_all_m_expand_index(lmax: int):
    expand_index = torch.zeros([(lmax + 1) ** 2]).long()
    for lval in range(lmax + 1):
        start_idx = lval**2
        length = 2 * lval + 1
        expand_index[start_idx : (start_idx + length)] = lval
    return expand_index


class EquivariantLayerNormArray(nn.Module):
    def __init__(
        self,
        lmax: int,
        num_channels: int,
        eps: float = 1e-5,
        affine: bool = True,
        normalization: str = "component",
    ):
        super().__init__()

        self.lmax = lmax
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine

        if affine:
            self.affine_weight = nn.Parameter(torch.ones(lmax + 1, num_channels))
            self.affine_bias = nn.Parameter(torch.zeros(num_channels))
        else:
            self.register_parameter("affine_weight", None)
            self.register_parameter("affine_bias", None)

        assert normalization in ["norm", "component"]
        self.normalization = normalization

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(lmax={self.lmax}, num_channels={self.num_channels}, eps={self.eps})"

    @torch.autocast("cuda", enabled=False)
    def forward(self, node_input):
        """
        Assume input is of shape [N, sphere_basis, C]
        """

        out = []

        for lval in range(self.lmax + 1):
            start_idx = lval**2
            length = 2 * lval + 1

            feature = node_input.narrow(1, start_idx, length)

            # For scalars, first compute and subtract the mean
            if lval == 0:
                feature_mean = torch.mean(feature, dim=2, keepdim=True)
                feature = feature - feature_mean

            # Then compute the rescaling factor (norm of each feature vector)
            # Rescaling of the norms themselves based on the option "normalization"
            if self.normalization == "norm":
                feature_norm = feature.pow(2).sum(dim=1, keepdim=True)  # [N, 1, C]
            elif self.normalization == "component":
                feature_norm = feature.pow(2).mean(dim=1, keepdim=True)  # [N, 1, C]

            feature_norm = torch.mean(feature_norm, dim=2, keepdim=True)  # [N, 1, 1]
            feature_norm = (feature_norm + self.eps).pow(-0.5)

            if self.affine:
                weight = self.affine_weight.narrow(0, lval, 1)  # [1, C]
                weight = weight.view(1, 1, -1)  # [1, 1, C]
                feature_norm = feature_norm * weight  # [N, 1, C]

            feature = feature * feature_norm

            if self.affine and lval == 0:
                bias = self.affine_bias
                bias = bias.view(1, 1, -1)
                feature = feature + bias

            out.append(feature)

        return torch.cat(out, dim=1)


class EquivariantLayerNormArraySphericalHarmonics(nn.Module):
    """
    1. Normalize over L = 0.
    2. Normalize across all m components from degrees L > 0.
    3. Do not normalize separately for different L (L > 0).
    """

    def __init__(
        self,
        lmax: int,
        num_channels: int,
        eps: float = 1e-5,
        affine: bool = True,
        normalization: str = "component",
        std_balance_degrees: bool = True,
    ):
        super().__init__()

        self.lmax = lmax
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine
        self.std_balance_degrees = std_balance_degrees

        # for L = 0
        self.norm_l0 = torch.nn.LayerNorm(
            self.num_channels, eps=self.eps, elementwise_affine=self.affine
        )

        # for L > 0
        if self.affine:
            self.affine_weight = nn.Parameter(torch.ones(self.lmax, self.num_channels))
        else:
            self.register_parameter("affine_weight", None)

        assert normalization in ["norm", "component"]
        self.normalization = normalization

        if self.std_balance_degrees:
            balance_degree_weight = torch.zeros((self.lmax + 1) ** 2 - 1, 1)
            for lval in range(1, self.lmax + 1):
                start_idx = lval**2 - 1
                length = 2 * lval + 1
                balance_degree_weight[start_idx : (start_idx + length), :] = (
                    1.0 / length
                )
            balance_degree_weight = balance_degree_weight / self.lmax
            self.register_buffer(
                "balance_degree_weight", balance_degree_weight, persistent=False
            )
        else:
            self.balance_degree_weight = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(lmax={self.lmax}, num_channels={self.num_channels}, eps={self.eps}, std_balance_degrees={self.std_balance_degrees})"

    @torch.autocast("cuda", enabled=False)
    def forward(self, node_input):
        """
        Assume input is of shape [N, sphere_basis, C]
        """

        out = []

        # for L = 0
        feature = node_input.narrow(1, 0, 1)
        feature = self.norm_l0(feature)
        out.append(feature)

        # for L > 0
        if self.lmax > 0:
            num_m_components = (self.lmax + 1) ** 2
            feature = node_input.narrow(1, 1, num_m_components - 1)

            # Then compute the rescaling factor (norm of each feature vector)
            # Rescaling of the norms themselves based on the option "normalization"
            if self.normalization == "norm":
                feature_norm = feature.pow(2).sum(dim=1, keepdim=True)  # [N, 1, C]
            elif self.normalization == "component":
                if self.std_balance_degrees:
                    feature_norm = feature.pow(
                        2
                    )  # [N, (L_max + 1)**2 - 1, C], without L = 0
                    feature_norm = torch.einsum(
                        "nic, ia -> nac",
                        feature_norm,
                        self.balance_degree_weight,
                    )  # [N, 1, C]
                else:
                    feature_norm = feature.pow(2).mean(dim=1, keepdim=True)  # [N, 1, C]

            feature_norm = torch.mean(feature_norm, dim=2, keepdim=True)  # [N, 1, 1]
            feature_norm = (feature_norm + self.eps).pow(-0.5)

            for lval in range(1, self.lmax + 1):
                start_idx = lval**2
                length = 2 * lval + 1
                feature = node_input.narrow(1, start_idx, length)  # [N, (2L + 1), C]
                if self.affine:
                    weight = self.affine_weight.narrow(0, (lval - 1), 1)  # [1, C]
                    weight = weight.view(1, 1, -1)  # [1, 1, C]
                    feature_scale = feature_norm * weight  # [N, 1, C]
                else:
                    feature_scale = feature_norm
                feature = feature * feature_scale
                out.append(feature)

        return torch.cat(out, dim=1)


class EquivariantRMSNormArraySphericalHarmonics(nn.Module):
    """
    1. Normalize across all m components from degrees L >= 0.
    """

    def __init__(
        self,
        lmax: int,
        num_channels: int,
        eps: float = 1e-5,
        affine: bool = True,
        normalization: str = "component",
    ):
        super().__init__()

        self.lmax = lmax
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine

        # for L >= 0
        if self.affine:
            self.affine_weight = nn.Parameter(
                torch.ones((self.lmax + 1), self.num_channels)
            )
        else:
            self.register_parameter("affine_weight", None)

        assert normalization in ["norm", "component"]
        self.normalization = normalization

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(lmax={self.lmax}, num_channels={self.num_channels}, eps={self.eps})"

    @torch.autocast("cuda", enabled=False)
    def forward(self, node_input):
        """
        Assume input is of shape [N, sphere_basis, C]
        """

        out = []

        # for L >= 0
        feature = node_input
        if self.normalization == "norm":
            feature_norm = feature.pow(2).sum(dim=1, keepdim=True)  # [N, 1, C]
        elif self.normalization == "component":
            feature_norm = feature.pow(2).mean(dim=1, keepdim=True)  # [N, 1, C]

        feature_norm = torch.mean(feature_norm, dim=2, keepdim=True)  # [N, 1, 1]
        feature_norm = (feature_norm + self.eps).pow(-0.5)

        for lval in range(self.lmax + 1):
            start_idx = lval**2
            length = 2 * lval + 1
            feature = node_input.narrow(1, start_idx, length)  # [N, (2L + 1), C]
            if self.affine:
                weight = self.affine_weight.narrow(0, lval, 1)  # [1, C]
                weight = weight.view(1, 1, -1)  # [1, 1, C]
                feature_scale = feature_norm * weight  # [N, 1, C]
            else:
                feature_scale = feature_norm
            feature = feature * feature_scale
            out.append(feature)

        return torch.cat(out, dim=1)


class EquivariantRMSNormArraySphericalHarmonicsV2(nn.Module):
    """
    1. Normalize across all m components from degrees L >= 0.
    2. Expand weights and multiply with normalized feature to prevent slicing and concatenation.
    """

    def __init__(
        self,
        lmax: int,
        num_channels: int,
        eps: float = 1e-5,
        affine: bool = True,
        normalization: str = "component",
        centering: bool = True,
        std_balance_degrees: bool = True,
    ):
        super().__init__()

        self.lmax = lmax
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine
        self.centering = centering
        self.std_balance_degrees = std_balance_degrees

        # for L >= 0
        if self.affine:
            self.affine_weight = nn.Parameter(
                torch.ones((self.lmax + 1), self.num_channels)
            )
            if self.centering:
                self.affine_bias = nn.Parameter(torch.zeros(self.num_channels))
            else:
                self.register_parameter("affine_bias", None)
        else:
            self.register_parameter("affine_weight", None)
            self.register_parameter("affine_bias", None)

        assert normalization in ["norm", "component"]
        self.normalization = normalization

        expand_index = get_l_to_all_m_expand_index(self.lmax)
        self.register_buffer("expand_index", expand_index, persistent=False)

        if self.std_balance_degrees:
            balance_degree_weight = torch.zeros((self.lmax + 1) ** 2, 1)
            for lval in range(self.lmax + 1):
                start_idx = lval**2
                length = 2 * lval + 1
                balance_degree_weight[start_idx : (start_idx + length), :] = (
                    1.0 / length
                )
            balance_degree_weight = balance_degree_weight / (self.lmax + 1)
            self.register_buffer(
                "balance_degree_weight", balance_degree_weight, persistent=False
            )
        else:
            self.balance_degree_weight = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(lmax={self.lmax}, num_channels={self.num_channels}, eps={self.eps}, centering={self.centering}, std_balance_degrees={self.std_balance_degrees})"

    @torch.autocast("cuda", enabled=False)
    def forward(self, node_input):
        """
        Assume input is of shape [N, sphere_basis, C]
        """

        feature = node_input

        if self.centering:
            feature_l0 = feature.narrow(1, 0, 1)
            feature_l0_mean = feature_l0.mean(dim=2, keepdim=True)  # [N, 1, 1]
            feature_l0 = feature_l0 - feature_l0_mean
            feature = torch.cat(
                (feature_l0, feature.narrow(1, 1, feature.shape[1] - 1)), dim=1
            )

        # for L >= 0
        if self.normalization == "norm":
            assert not self.std_balance_degrees
            feature_norm = feature.pow(2).sum(dim=1, keepdim=True)  # [N, 1, C]
        elif self.normalization == "component":
            if self.std_balance_degrees:
                feature_norm = feature.pow(2)  # [N, (L_max + 1)**2, C]
                feature_norm = torch.einsum(
                    "nic, ia -> nac", feature_norm, self.balance_degree_weight
                )  # [N, 1, C]
            else:
                feature_norm = feature.pow(2).mean(dim=1, keepdim=True)  # [N, 1, C]

        feature_norm = torch.mean(feature_norm, dim=2, keepdim=True)  # [N, 1, 1]
        feature_norm = (feature_norm + self.eps).pow(-0.5)

        if self.affine:
            weight = self.affine_weight.view(
                1, (self.lmax + 1), self.num_channels
            )  # [1, L_max + 1, C]
            weight = torch.index_select(
                weight, dim=1, index=self.expand_index
            )  # [1, (L_max + 1)**2, C]
            feature_norm = feature_norm * weight  # [N, (L_max + 1)**2, C]

        out = feature * feature_norm

        if self.affine and self.centering:
            out[:, 0:1, :] = out.narrow(1, 0, 1) + self.affine_bias.view(
                1, 1, self.num_channels
            )

        return out


class EquivariantDegreeLayerScale(nn.Module):
    """
    1. Similar to Layer Scale used in CaiT (Going Deeper With Image Transformers (ICCV'21)), we scale the output of both attention and FFN.
    2. For degree L > 0, we scale down the square root of 2 * L, which is to emulate halving the number of channels when using higher L.
    """

    def __init__(self, lmax: int, num_channels: int, scale_factor: float = 2.0) -> None:
        super().__init__()

        self.lmax = lmax
        self.num_channels = num_channels
        self.scale_factor = scale_factor

        self.affine_weight = nn.Parameter(
            torch.ones(1, (self.lmax + 1), self.num_channels)
        )
        for lval in range(1, self.lmax + 1):
            self.affine_weight.data[0, lval, :].mul_(
                1.0 / math.sqrt(self.scale_factor * lval)
            )
        expand_index = get_l_to_all_m_expand_index(self.lmax)
        self.register_buffer("expand_index", expand_index, persistent=False)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(lmax={self.lmax}, num_channels={self.num_channels}, scale_factor={self.scale_factor})"

    def forward(self, node_input):
        weight = torch.index_select(
            self.affine_weight, dim=1, index=self.expand_index
        )  # [1, (L_max + 1)**2, C]
        return node_input * weight  # [N, (L_max + 1)**2, C]



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import math

import numpy as np
import torch


def sum_partitions(x: torch.Tensor, partition_idxs: torch.Tensor) -> torch.Tensor:
    sums = torch.zeros(partition_idxs.shape[0] - 1, device=x.device, dtype=x.dtype)
    for idx in range(partition_idxs.shape[0] - 1):
        sums[idx] = x[partition_idxs[idx] : partition_idxs[idx + 1]].sum()
    return sums


def get_counts(x: torch.Tensor, length: int):
    dtype = x.dtype
    device = x.device
    return torch.zeros(length, device=device, dtype=dtype).scatter_reduce(
        dim=0,
        index=x,
        src=torch.ones(x.shape[0], device=device, dtype=dtype),
        reduce="sum",
    )


def compute_neighbors(data, edge_index):
    # Get number of neighbors
    num_neighbors = get_counts(edge_index[1], data.natoms.sum())

    # Get number of neighbors per image
    image_indptr = torch.zeros(
        data.natoms.shape[0] + 1, device=data.pos.device, dtype=torch.long
    )
    image_indptr[1:] = torch.cumsum(data.natoms, dim=0)
    return sum_partitions(num_neighbors, image_indptr)


def get_max_neighbors_mask(
    natoms,
    index,
    atom_distance,
    max_num_neighbors_threshold,
    degeneracy_tolerance: float = 0.01,
    enforce_max_strictly: bool = False,
):
    """
    Give a mask that filters out edges so that each atom has at most
    `max_num_neighbors_threshold` neighbors.
    Assumes that `index` is sorted.

    Enforcing the max strictly can force the arbitrary choice between
    degenerate edges. This can lead to undesired behaviors; for
    example, bulk formation energies which are not invariant to
    unit cell choice.

    A degeneracy tolerance can help prevent sudden changes in edge
    existence from small changes in atom position, for example,
    rounding errors, slab relaxation, temperature, etc.
    """

    device = natoms.device
    num_atoms = natoms.sum()

    # Get number of neighbors
    num_neighbors = get_counts(index, num_atoms)

    max_num_neighbors = num_neighbors.max()
    if max_num_neighbors_threshold > 0:
        num_neighbors_thresholded = num_neighbors.clamp(max=max_num_neighbors_threshold)
    else:
        num_neighbors_thresholded = num_neighbors

    # Get number of (thresholded) neighbors per image
    image_indptr = torch.zeros(natoms.shape[0] + 1, device=device, dtype=torch.long)
    image_indptr[1:] = torch.cumsum(natoms, dim=0)
    num_neighbors_image = sum_partitions(num_neighbors_thresholded, image_indptr)

    # If max_num_neighbors is below the threshold, return early
    if (
        max_num_neighbors <= max_num_neighbors_threshold
        or max_num_neighbors_threshold <= 0
    ):
        mask_num_neighbors = torch.tensor([True], dtype=bool, device=device).expand_as(
            index
        )
        return mask_num_neighbors, num_neighbors_image

    # Create a tensor of size [num_atoms, max_num_neighbors] to sort the distances of the neighbors.
    # Fill with infinity so we can easily remove unused distances later.
    distance_sort = torch.full(
        [num_atoms * max_num_neighbors],
        np.inf,
        device=device,
        dtype=atom_distance.dtype,
    )

    # Create an index map to map distances from atom_distance to distance_sort
    # index_sort_map assumes index to be sorted
    index_neighbor_offset = torch.cumsum(num_neighbors, dim=0) - num_neighbors
    index_neighbor_offset_expand = torch.repeat_interleave(
        index_neighbor_offset, num_neighbors
    )
    index_sort_map = (
        index * max_num_neighbors
        + torch.arange(len(index), device=device)
        - index_neighbor_offset_expand
    )
    distance_sort.index_copy_(0, index_sort_map, atom_distance)
    distance_sort = distance_sort.view(num_atoms, max_num_neighbors)

    # Sort neighboring atoms based on distance
    distance_sort, index_sort = torch.sort(distance_sort, dim=1)

    # Select the max_num_neighbors_threshold neighbors that are closest
    if enforce_max_strictly:
        distance_sort = distance_sort[:, :max_num_neighbors_threshold]
        index_sort = index_sort[:, :max_num_neighbors_threshold]
        max_num_included = max_num_neighbors_threshold

    else:
        effective_cutoff = (
            distance_sort[:, max_num_neighbors_threshold] + degeneracy_tolerance
        )
        is_included = torch.le(distance_sort.T, effective_cutoff)

        # Set all undesired edges to infinite length to be removed later
        distance_sort[~is_included.T] = np.inf

        # Subselect tensors for efficiency
        num_included_per_atom = torch.sum(is_included, dim=0)
        max_num_included = torch.max(num_included_per_atom)
        distance_sort = distance_sort[:, :max_num_included]
        index_sort = index_sort[:, :max_num_included]

        # Recompute the number of neighbors
        num_neighbors_thresholded = num_neighbors.clamp(max=num_included_per_atom)

        num_neighbors_image = sum_partitions(num_neighbors_thresholded, image_indptr)

    # Offset index_sort so that it indexes into index
    index_sort = index_sort + index_neighbor_offset.view(-1, 1).expand(
        -1, max_num_included
    )
    # Remove "unused pairs" with infinite distances
    mask_finite = torch.isfinite(distance_sort)
    index_sort = torch.masked_select(index_sort, mask_finite)

    # At this point index_sort contains the index into index of the
    # closest max_num_neighbors_threshold neighbors per atom
    # Create a mask to remove all pairs not in index_sort
    mask_num_neighbors = torch.zeros(len(index), device=device, dtype=bool)
    mask_num_neighbors.index_fill_(0, index_sort, True)
    return mask_num_neighbors, num_neighbors_image


# @torch.no_grad() # breaks torch compile test with OOM
def radius_graph_pbc(
    data,
    radius,
    max_num_neighbors_threshold,
    enforce_max_neighbors_strictly: bool = False,
    pbc: torch.Tensor | None = None,
):
    pbc = canonical_pbc(data, pbc)

    device = data.pos.device
    batch_size = len(data.natoms)

    # position of the atoms
    atom_pos = data.pos

    # Before computing the pairwise distances between atoms, first create a list of atom indices to compare for the entire batch
    num_atoms_per_image = data.natoms
    num_atoms_per_image_sqr = (num_atoms_per_image**2).long()

    # index offset between images
    index_offset = torch.cumsum(num_atoms_per_image, dim=0) - num_atoms_per_image

    index_offset_expand = torch.repeat_interleave(index_offset, num_atoms_per_image_sqr)
    num_atoms_per_image_expand = torch.repeat_interleave(
        num_atoms_per_image, num_atoms_per_image_sqr
    )

    # Compute a tensor containing sequences of numbers that range from 0 to num_atoms_per_image_sqr for each image
    # that is used to compute indices for the pairs of atoms. This is a very convoluted way to implement
    # the following (but 10x faster since it removes the for loop)
    # for batch_idx in range(batch_size):
    #    batch_count = torch.cat([batch_count, torch.arange(num_atoms_per_image_sqr[batch_idx], device=device)], dim=0)
    num_atom_pairs = torch.sum(num_atoms_per_image_sqr)
    index_sqr_offset = (
        torch.cumsum(num_atoms_per_image_sqr, dim=0) - num_atoms_per_image_sqr
    )
    index_sqr_offset = torch.repeat_interleave(
        index_sqr_offset, num_atoms_per_image_sqr
    )
    atom_count_sqr = torch.arange(num_atom_pairs, device=device) - index_sqr_offset

    # Compute the indices for the pairs of atoms (using division and mod)
    # If the systems get too large this apporach could run into numerical precision issues
    index1 = (
        torch.div(atom_count_sqr, num_atoms_per_image_expand, rounding_mode="floor")
    ) + index_offset_expand
    index2 = (atom_count_sqr % num_atoms_per_image_expand) + index_offset_expand
    # Get the positions for each atom
    pos1 = torch.index_select(atom_pos, 0, index1)
    pos2 = torch.index_select(atom_pos, 0, index2)

    # Calculate required number of unit cells in each direction.
    # Smallest distance between planes separated by a1 is
    # 1 / ||(a2 x a3) / V||_2, since a2 x a3 is the area of the plane.
    # Note that the unit cell volume V = a1 * (a2 x a3) and that
    # (a2 x a3) / V is also the reciprocal primitive vector
    # (crystallographer's definition).

    cross_a2a3 = torch.cross(data.cell[:, 1], data.cell[:, 2], dim=-1)
    cell_vol = torch.sum(data.cell[:, 0] * cross_a2a3, dim=-1, keepdim=True)

    if pbc[0]:
        inv_min_dist_a1 = torch.norm(cross_a2a3 / cell_vol, p=2, dim=-1)
        rep_a1 = torch.ceil(radius * inv_min_dist_a1)
    else:
        rep_a1 = data.cell.new_zeros(1)

    if pbc[1]:
        cross_a3a1 = torch.cross(data.cell[:, 2], data.cell[:, 0], dim=-1)
        inv_min_dist_a2 = torch.norm(cross_a3a1 / cell_vol, p=2, dim=-1)
        rep_a2 = torch.ceil(radius * inv_min_dist_a2)
    else:
        rep_a2 = data.cell.new_zeros(1)

    if pbc[2]:
        cross_a1a2 = torch.cross(data.cell[:, 0], data.cell[:, 1], dim=-1)
        inv_min_dist_a3 = torch.norm(cross_a1a2 / cell_vol, p=2, dim=-1)
        rep_a3 = torch.ceil(radius * inv_min_dist_a3)
    else:
        rep_a3 = data.cell.new_zeros(1)

    # Take the max over all images for uniformity. This is essentially padding.
    # Note that this can significantly increase the number of computed distances
    # if the required repetitions are very different between images
    # (which they usually are). Changing this to sparse (scatter) operations
    # might be worth the effort if this function becomes a bottleneck.
    max_rep = [rep_a1.max(), rep_a2.max(), rep_a3.max()]

    # Tensor of unit cells
    cells_per_dim = [
        torch.arange(-rep.item(), rep.item() + 1, device=device, dtype=data.cell.dtype)
        for rep in max_rep
    ]
    unit_cell = torch.cartesian_prod(*cells_per_dim)
    num_cells = len(unit_cell)
    unit_cell_per_atom = unit_cell.view(1, num_cells, 3).repeat(len(index2), 1, 1)
    unit_cell = torch.transpose(unit_cell, 0, 1)
    unit_cell_batch = unit_cell.view(1, 3, num_cells).expand(batch_size, -1, -1)

    # Compute the x, y, z positional offsets for each cell in each image
    data_cell = torch.transpose(data.cell, 1, 2)
    pbc_offsets = torch.bmm(data_cell, unit_cell_batch)
    pbc_offsets_per_atom = torch.repeat_interleave(
        pbc_offsets, num_atoms_per_image_sqr, dim=0
    )

    # Expand the positions and indices for the 9 cells
    pos1 = pos1.view(-1, 3, 1).expand(-1, -1, num_cells)
    pos2 = pos2.view(-1, 3, 1).expand(-1, -1, num_cells)
    index1 = index1.view(-1, 1).repeat(1, num_cells).view(-1)
    index2 = index2.view(-1, 1).repeat(1, num_cells).view(-1)
    # Add the PBC offsets for the second atom
    pos2 = pos2 + pbc_offsets_per_atom

    # Compute the squared distance between atoms
    atom_distance_sqr = torch.sum((pos1 - pos2) ** 2, dim=1)
    atom_distance_sqr = atom_distance_sqr.view(-1)

    # Remove pairs that are too far apart
    mask_within_radius = torch.le(atom_distance_sqr, radius * radius)
    # Remove pairs with the same atoms (distance = 0.0)
    mask_not_same = torch.gt(atom_distance_sqr, 0.0001)
    mask = torch.logical_and(mask_within_radius, mask_not_same)
    index1 = torch.masked_select(index1, mask)
    index2 = torch.masked_select(index2, mask)
    unit_cell = torch.masked_select(
        unit_cell_per_atom.view(-1, 3), mask.view(-1, 1).expand(-1, 3)
    )
    unit_cell = unit_cell.view(-1, 3)
    atom_distance_sqr = torch.masked_select(atom_distance_sqr, mask)

    mask_num_neighbors, num_neighbors_image = get_max_neighbors_mask(
        natoms=data.natoms,
        index=index1,
        atom_distance=atom_distance_sqr,
        max_num_neighbors_threshold=max_num_neighbors_threshold,
        enforce_max_strictly=enforce_max_neighbors_strictly,
    )

    if not torch.all(mask_num_neighbors):
        # Mask out the atoms to ensure each atom has at most max_num_neighbors_threshold neighbors
        index1 = torch.masked_select(index1, mask_num_neighbors)
        index2 = torch.masked_select(index2, mask_num_neighbors)
        unit_cell = torch.masked_select(
            unit_cell.view(-1, 3), mask_num_neighbors.view(-1, 1).expand(-1, 3)
        )
        unit_cell = unit_cell.view(-1, 3)

    edge_index = torch.stack((index2, index1))

    return edge_index, unit_cell, num_neighbors_image


def canonical_pbc(data, pbc: torch.Tensor | None):
    assert hasattr(data, "pbc"), "AtomicData does not have pbc set"
    if pbc is None and hasattr(data, "pbc"):
        data.pbc = torch.atleast_2d(data.pbc)
        pbc = torch.BoolTensor([True, True, True])
        for i in range(3):
            if not torch.any(data.pbc[:, i]).item():
                pbc[i] = False
            elif torch.all(data.pbc[:, i]).item():
                pbc[i] = True
            else:
                raise RuntimeError(
                    "Different structures in the batch have different PBC configurations. This is not currently supported."
                )
    # elif pbc is not None and hasattr(data, "pbc"):
    #     # This can be on a different device, deffering to a new PR to fix this TODO
    #     if (pbc != data.pbc).all():
    #         logging.warning("PBC provided to radius_graph_pbc differs from data.pbc")
    elif pbc is None:
        pbc = torch.BoolTensor([True, True, True])

    assert isinstance(pbc, torch.Tensor)
    assert pbc.ndim == 1
    assert pbc.shape[0] == 3
    return list(pbc)


def box_size_warning(cell, pos, pbc):
    if hasattr(box_size_warning, "already_printed"):
        return
    box_size_warning.already_printed = True
    logging.warning(
        f"PBCv2: graph generation encountered a very large box. The size of the cell is {cell} and min/max positions are {pos.min(),pos.max()}. Performance will be slower than optimal."
    )
    if any(pbc):
        logging.warning(
            "PBCv2: Does this system require PBC=True or will PBC=False work?"
        )


@torch.no_grad()
def radius_graph_pbc_v2(
    data,
    radius,
    max_num_neighbors_threshold,
    enforce_max_neighbors_strictly: bool = False,
    pbc: torch.Tensor | None = None,
):
    pbc = canonical_pbc(data, pbc)

    device = data.pos.device
    batch_size = len(data.natoms)
    data_batch_idxs = (
        data.batch
        if data.batch is not None
        else data.pos.new_zeros(data.natoms[0], dtype=torch.int)
    )
    # Resolution of the grid cells, should be less than the radius
    grid_resolution = radius / 1.99

    # This function assumes that all atoms are within the unti cell. If atoms
    # are outside of the unit cell, it will not work correctly.
    # position of the atoms
    atom_pos = data.pos
    num_atoms = len(data.pos)
    num_atoms_per_image = data.natoms

    # Calculate required number of unit cells in each direction.
    # Smallest distance between planes separated by a1 is
    # 1 / ||(a2 x a3) / V||_2, since a2 x a3 is the area of the plane.
    # Note that the unit cell volume V = a1 * (a2 x a3) and that
    # (a2 x a3) / V is also the reciprocal primitive vector
    # (crystallographer's definition).

    cross_a2a3 = torch.cross(data.cell[:, 1], data.cell[:, 2], dim=-1)
    cell_vol = torch.sum(data.cell[:, 0] * cross_a2a3, dim=-1, keepdim=True)

    if pbc[0]:
        inv_min_dist_a1 = torch.norm(cross_a2a3 / cell_vol, p=2, dim=-1)
        rep_a1 = torch.ceil(radius * inv_min_dist_a1)
    else:
        rep_a1 = data.cell.new_zeros(batch_size)

    if pbc[1]:
        cross_a3a1 = torch.cross(data.cell[:, 2], data.cell[:, 0], dim=-1)
        inv_min_dist_a2 = torch.norm(cross_a3a1 / cell_vol, p=2, dim=-1)
        rep_a2 = torch.ceil(radius * inv_min_dist_a2)
    else:
        rep_a2 = data.cell.new_zeros(batch_size)

    if pbc[2]:
        cross_a1a2 = torch.cross(data.cell[:, 0], data.cell[:, 1], dim=-1)
        inv_min_dist_a3 = torch.norm(cross_a1a2 / cell_vol, p=2, dim=-1)
        rep_a3 = torch.ceil(radius * inv_min_dist_a3)
    else:
        rep_a3 = data.cell.new_zeros(batch_size)

    rep = torch.cat([rep_a1.view(-1, 1), rep_a2.view(-1, 1), rep_a3.view(-1, 1)], dim=1)
    cells_per_image = (
        (rep[:, 0] * 2 + 1.0) * (rep[:, 1] * 2 + 1.0) * (rep[:, 2] * 2 + 1.0)
    ).long()

    # Create a tensor of unit cells for each image
    unit_cell = torch.zeros(
        torch.sum(cells_per_image), 3, device=device, dtype=data.cell.dtype
    )
    offset = 0
    for i in range(batch_size):
        cells_x = torch.arange(
            -rep[i][0], rep[i][0] + 1, device=device, dtype=torch.float
        )
        cells_y = torch.arange(
            -rep[i][1], rep[i][1] + 1, device=device, dtype=torch.float
        )
        cells_z = torch.arange(
            -rep[i][2], rep[i][2] + 1, device=device, dtype=torch.float
        )
        unit_cell[offset : cells_per_image[i] + offset] = torch.cartesian_prod(
            cells_x, cells_y, cells_z
        )
        offset = offset + cells_per_image[i]

    # Compute the x, y, z positional offsets for each cell in each image
    cell_matrix = torch.transpose(data.cell, 1, 2)
    cell_matrix = torch.repeat_interleave(cell_matrix, cells_per_image, dim=0)
    pbc_cell_offsets = torch.bmm(cell_matrix, unit_cell.view(-1, 3, 1)).squeeze(-1)

    # Position of the target atoms for the edges
    target_atom_pos = atom_pos
    target_atom_image = data_batch_idxs

    # Compute the position of the source atoms for the edges. There are
    # more source atoms than target atoms, since the source atoms are
    # tiled by the PBC cells.
    num_cells_per_atom = torch.repeat_interleave(cells_per_image, num_atoms_per_image)
    source_atom_index = torch.repeat_interleave(
        torch.arange(num_atoms, device=device).long(), num_cells_per_atom
    )
    source_atom_image = data_batch_idxs[source_atom_index]
    source_atom_pos = atom_pos[source_atom_index]

    # For each atom the index of the PBC cell
    pbc_cell_index = torch.tensor([], device=device).long()
    offset = 0
    for i in range(batch_size):
        cell_indices = (
            torch.arange(offset, offset + cells_per_image[i], device=device)
            .repeat(num_atoms_per_image[i])
            .long()
        )
        pbc_cell_index = torch.cat([pbc_cell_index, cell_indices], dim=0)
        offset = offset + cells_per_image[i]

    # Remember the source cell for later use
    source_cell = unit_cell[pbc_cell_index]

    # Compute their PBC cell offsets
    source_pbc_cell_offsets = pbc_cell_offsets[pbc_cell_index]
    # Add on the PBC cell offsets
    source_atom_pos = source_atom_pos + source_pbc_cell_offsets

    # Given the positions of all the atoms has been computed, we
    # split them up using a cubic grid to make pairwise comparisons
    # more computationally efficient. The resolution of the grid
    # is grid_resolution.

    # Compute the grid index for each dimension for each atom
    max_internal_cell = (
        max(source_atom_pos.abs().max(), target_atom_pos.abs().max()) / grid_resolution
    )
    if max_internal_cell > 200:
        box_size_warning(data.cell, data.pos, pbc)
        grid_resolution = max(
            grid_resolution,
            max(source_atom_pos.abs().max(), target_atom_pos.abs().max()) / 200,
        )
    source_atom_grid = torch.floor(source_atom_pos / grid_resolution).long()
    target_atom_grid = torch.floor(target_atom_pos / grid_resolution).long()

    # Find the min and max grid index for each image
    unique_atom_image, num_source_atoms_per_image = torch.unique(
        source_atom_image, return_counts=True
    )
    max_num_source_atoms = torch.max(num_source_atoms_per_image)
    # Create a new array with size [batch_size, max_num_source_atoms] to hold
    # the grid indices for each atom. We can then perform max/min on each
    # image separately.
    # First, create a mapping from the array of source atoms to the 2D array.
    source_atom_offset_per_image = torch.cumsum(num_source_atoms_per_image, dim=0).roll(
        1
    )
    source_atom_offset_per_image[0] = 0
    source_atom_offset_per_image = torch.repeat_interleave(
        source_atom_offset_per_image, num_source_atoms_per_image, 0
    )
    source_atom_mapping = (
        torch.arange(len(source_atom_offset_per_image), device=device, dtype=torch.long)
        - source_atom_offset_per_image
    )
    source_atom_mapping = (
        source_atom_mapping
        + torch.repeat_interleave(unique_atom_image, num_source_atoms_per_image, 0)
        * max_num_source_atoms
    )

    # Create 2D array
    source_atom_grid_per_image = torch.zeros(
        batch_size * max_num_source_atoms, 3, device=device, dtype=torch.long
    )
    # Populate with the grid values
    source_atom_grid_per_image[source_atom_mapping] = source_atom_grid
    source_atom_grid_per_image = source_atom_grid_per_image.view(
        batch_size, max_num_source_atoms, 3
    )
    # Perform min and max operations per image
    grid_min, no_op = torch.min(source_atom_grid_per_image, dim=1)
    grid_max, no_op = torch.max(source_atom_grid_per_image, dim=1)

    # Size of grid in each dimension for each image
    grid_size = grid_max - grid_min + 1
    grid_length = grid_size[:, 0] * grid_size[:, 1] * grid_size[:, 2]
    # Offset between grids for each image
    grid_offset = torch.cat(
        [torch.tensor([0], device=device), torch.cumsum(grid_length, dim=0)], dim=0
    )

    num_grid_cells = torch.sum(grid_length)

    # Subtract the minimum grid index so they are zero indexed
    source_atom_grid = source_atom_grid - grid_min[source_atom_image]
    target_atom_grid = target_atom_grid - grid_min[target_atom_image]

    # Compute the grid id which is a combination of the grid indices in the x,y,z directions
    # Grid id is x + y*grid_size[0] + z*grid_size[0]*grid_size[1] + offset
    source_atom_grid_size = grid_size[source_atom_image]
    source_atom_grid_offset = grid_offset[source_atom_image]
    source_atom_grid_id = (
        source_atom_grid[:, 0]
        + source_atom_grid[:, 1] * source_atom_grid_size[:, 0]
        + source_atom_grid[:, 2]
        * source_atom_grid_size[:, 0]
        * source_atom_grid_size[:, 1]
    )
    source_atom_grid_id = source_atom_grid_id + source_atom_grid_offset

    target_atom_grid_size = grid_size[target_atom_image]
    target_atom_grid_offset = grid_offset[target_atom_image]
    target_atom_grid_id = (
        target_atom_grid[:, 0]
        + target_atom_grid[:, 1] * target_atom_grid_size[:, 0]
        + target_atom_grid[:, 2]
        * target_atom_grid_size[:, 0]
        * target_atom_grid_size[:, 1]
    )
    target_atom_grid_id = target_atom_grid_id + target_atom_grid_offset

    # Compute a mapping from the array of atoms to a 2D array containing
    # all the atoms in each grid cell of size [num_grid_cells, max_atoms_per_grid_cell].
    # Pad each list of atoms for each grid cell so each
    # list is of the same length - max_atoms_per_grid_cell.
    sort_grid_id, sort_indices = torch.sort(source_atom_grid_id)
    grid_cell_atom_count = torch.zeros(
        num_grid_cells, device=device, dtype=source_atom_grid_id.dtype
    )
    grid_cell_atom_count.index_add_(
        0, source_atom_grid_id, torch.ones_like(source_atom_grid_id)
    )

    # Maximum number of atoms in a grid cell used to pad the array of atoms in each
    max_atoms_per_grid_cell = torch.max(grid_cell_atom_count)

    # Compute a mapping from the grid cell lists to the atoms in that grid cell
    cum_sum_grid_cell_atom_count = torch.cumsum(grid_cell_atom_count, 0)
    cum_sum_grid_cell_atom_count = torch.roll(cum_sum_grid_cell_atom_count, 1, 0)
    cum_sum_grid_cell_atom_count[0] = 0
    cum_sum_grid_cell_offset = torch.repeat_interleave(
        cum_sum_grid_cell_atom_count, grid_cell_atom_count, dim=0
    )
    grid_cell_offset = (  # If this OOMs it could be because boxsize is large and PBC is on
        torch.arange(num_grid_cells, device=device) * max_atoms_per_grid_cell
    )
    grid_cell_offset = torch.repeat_interleave(
        grid_cell_offset, grid_cell_atom_count, dim=0
    )
    grid_atom_map = (
        torch.arange(len(grid_cell_offset), device=device)
        + grid_cell_offset
        - cum_sum_grid_cell_offset
    )

    # If an entry doesn't have an atom, set to a value of -1
    grid_atom_index = (
        torch.zeros(
            num_grid_cells * max_atoms_per_grid_cell, dtype=torch.long, device=device
        )
        - 1
    )
    # Populate the 2D array with the atom indices
    grid_atom_index[grid_atom_map] = sort_indices
    grid_atom_index = grid_atom_index.view(num_grid_cells, max_atoms_per_grid_cell)
    # Add a Null grid cell to the end
    grid_atom_index = torch.cat(
        [
            grid_atom_index,
            torch.zeros(max_atoms_per_grid_cell, device=device).view(1, -1).long() - 1,
        ],
        dim=0,
    )
    null_grid_index = num_grid_cells

    # How many grid cells do we need to include in each direction given the search radius?
    padding_size = math.floor(radius / grid_resolution) + 1
    num_padding_grid_cells = (
        (2 * padding_size + 1) * (2 * padding_size + 1) * (2 * padding_size + 1)
    )

    padding_offsets = torch.arange(
        -padding_size, padding_size + 1, device=device, dtype=torch.long
    )
    padding_offsets = padding_offsets.view(1, -1).repeat(3, 1)
    padding_offsets = torch.cartesian_prod(
        padding_offsets[0, :], padding_offsets[1, :], padding_offsets[2, :]
    )
    padding_offsets = padding_offsets.unsqueeze(0).repeat(batch_size, 1, 1)

    grid_index = target_atom_grid.view(-1, 1, 3).repeat(1, num_padding_grid_cells, 1)
    grid_index = grid_index + padding_offsets[target_atom_image]
    max_index = target_atom_grid_size.view(-1, 1, 3).expand(
        -1, num_padding_grid_cells, -1
    )
    out_of_bounds = torch.logical_and(grid_index.ge(0), grid_index.lt(max_index))
    out_of_bounds = torch.all(out_of_bounds, dim=2).ne(True)

    padding_offsets[:, :, 1] = padding_offsets[:, :, 1] * grid_size.unsqueeze(-1)[:, 0]
    padding_offsets[:, :, 2] = (
        padding_offsets[:, :, 2]
        * grid_size.unsqueeze(-1)[:, 0]
        * grid_size.unsqueeze(-1)[:, 1]
    )
    padding_offsets = torch.sum(padding_offsets, dim=2)

    # For every cell, compute a list of neighboring grid cells.
    neighboring_grid_cells = target_atom_grid_id.view(-1, 1).repeat(
        1, num_padding_grid_cells
    )
    neighboring_grid_cells = neighboring_grid_cells + padding_offsets[target_atom_image]

    # Filter neighboring cells that are out of bounds
    neighboring_grid_cells.masked_fill_(out_of_bounds, null_grid_index)

    target_atom_edge_index = (
        torch.arange(len(target_atom_pos), device=device)
        .view(-1, 1, 1)
        .repeat(1, num_padding_grid_cells, max_atoms_per_grid_cell)
    )
    source_atom_edge_index = grid_atom_index[neighboring_grid_cells]

    # Remove padded atoms
    atom_pad_mask = source_atom_edge_index.ne(-1)
    source_atom_edge_index = torch.masked_select(source_atom_edge_index, atom_pad_mask)
    target_atom_edge_index = torch.masked_select(target_atom_edge_index, atom_pad_mask)

    # Get the atom positions
    source_atom_edge_pos = source_atom_pos[source_atom_edge_index]
    target_atom_edge_pos = target_atom_pos[target_atom_edge_index]

    # Compute their distances
    atom_distance_sqr = torch.sum(
        (target_atom_edge_pos - source_atom_edge_pos) ** 2, dim=1
    )
    # Is the distance within the radius and not 0 (the same atom)
    within_radius_mask = torch.logical_and(
        atom_distance_sqr.le(radius * radius), atom_distance_sqr.ne(0.0)
    )
    source_atom_edge_index = torch.masked_select(
        source_atom_edge_index, within_radius_mask
    )
    target_atom_edge_index = torch.masked_select(
        target_atom_edge_index, within_radius_mask
    )
    atom_distance_sqr = torch.masked_select(atom_distance_sqr, within_radius_mask)

    # Get the return values
    # The indices of the atoms for each edge
    source_idx = source_atom_index[source_atom_edge_index]
    target_idx = target_atom_edge_index
    # The cell for each source atom
    source_cell = source_cell[source_atom_edge_index]
    # The number of edge per image
    no_op, num_neighbors_image = torch.unique(
        source_atom_image[source_atom_edge_index], return_counts=True
    )

    # Reduce the number of neighbors for each atom to the
    # desired threshold max_num_neighbors_threshold
    mask_num_neighbors, num_neighbors_image = get_max_neighbors_mask(
        natoms=data.natoms,
        index=target_idx,
        atom_distance=atom_distance_sqr,
        max_num_neighbors_threshold=max_num_neighbors_threshold,
        enforce_max_strictly=enforce_max_neighbors_strictly,
    )

    if not torch.all(mask_num_neighbors):
        # Mask out the atoms to ensure each atom has at most max_num_neighbors_threshold neighbors
        target_idx = torch.masked_select(target_idx, mask_num_neighbors)
        source_idx = torch.masked_select(source_idx, mask_num_neighbors)
        source_cell = torch.masked_select(
            source_cell.view(-1, 3), mask_num_neighbors.view(-1, 1).expand(-1, 3)
        )
        source_cell = source_cell.view(-1, 3)

    edge_index = torch.stack((source_idx, target_idx))

    return edge_index, source_cell, num_neighbors_image



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import math

import torch
import torch.nn as nn


@torch.jit.script
def gaussian(x: torch.Tensor, mean, std) -> torch.Tensor:
    a = (2 * math.pi) ** 0.5
    return torch.exp(-0.5 * (((x - mean) / std) ** 2)) / (a * std)


class PolynomialEnvelope(torch.nn.Module):
    """
    Polynomial envelope function that ensures a smooth cutoff.
    """

    def __init__(self, exponent: int = 5) -> None:
        super().__init__()
        assert exponent > 0
        self.p: float = float(exponent)
        self.a: float = -(self.p + 1) * (self.p + 2) / 2
        self.b: float = self.p * (self.p + 2)
        self.c: float = -self.p * (self.p + 1) / 2

    def forward(self, d_scaled: torch.Tensor) -> torch.Tensor:
        env_val = 1 + (d_scaled**self.p) * (
            self.a + d_scaled * (self.b + self.c * d_scaled)
        )
        return torch.where(d_scaled < 1, env_val, 0)


class GaussianSmearing(torch.nn.Module):
    def __init__(
        self,
        start: float = -5.0,
        stop: float = 5.0,
        num_gaussians: int = 50,
        basis_width_scalar: float = 1.0,
    ) -> None:
        super().__init__()
        self.num_output = num_gaussians
        offset = torch.linspace(start, stop, num_gaussians)
        self.coeff = -0.5 / (basis_width_scalar * (offset[1] - offset[0])).item() ** 2
        self.register_buffer("offset", offset, persistent=False)

    def forward(self, dist) -> torch.Tensor:
        dist = dist.view(-1, 1) - self.offset.view(1, -1)
        return torch.exp(self.coeff * torch.pow(dist, 2))


class RadialMLP(nn.Module):
    """
    Contruct a radial function (linear layers + layer normalization + SiLU) given a list of channels
    """

    def __init__(self, channels_list) -> None:
        super().__init__()
        modules = []
        input_channels = channels_list[0]
        for i in range(len(channels_list)):
            if i == 0:
                continue

            modules.append(nn.Linear(input_channels, channels_list[i], bias=True))
            input_channels = channels_list[i]

            if i == len(channels_list) - 1:
                break

            modules.append(nn.LayerNorm(channels_list[i]))
            modules.append(torch.nn.SiLU())

        self.net = nn.Sequential(*modules)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        return self.net(inputs)



from __future__ import annotations

import argparse
import logging
import os
from pathlib import Path

import yaml

from fairchem.core.scripts.create_finetune_dataset import (
    compute_normalizer_and_linear_reference,
    launch_processing,
)
from fairchem.core.units.mlip_unit.api.inference import UMATask

logging.basicConfig(level=logging.INFO)

TEMPLATE_DIR = Path("configs/uma/finetune")
DATA_YAML_DIR = Path("data")
REGRESSION_LABEL_TO_TASK_YAML = {
    "e": DATA_YAML_DIR / Path("uma_conserving_data_task_energy.yaml"),
    "ef": DATA_YAML_DIR / Path("uma_conserving_data_task_energy_force.yaml"),
    "efs": DATA_YAML_DIR / Path("uma_conserving_data_task_energy_force_stress.yaml"),
}
UMA_SM_FINETUNE_YAML = Path("uma_sm_finetune_template.yaml")


def create_yaml(
    train_path: str,
    val_path: str,
    force_rms: float,
    linref_coeff: list,
    output_dir: str,
    dataset_name: str,
    regression_tasks: str,
    base_model_name: str,
):
    data_task_yaml = TEMPLATE_DIR / REGRESSION_LABEL_TO_TASK_YAML[regression_tasks]
    with open(data_task_yaml) as file:
        template = yaml.safe_load(file)
        template["dataset_name"] = dataset_name
        template["normalizer_rmsd"] = force_rms
        template["elem_refs"] = linref_coeff
        template["train_dataset"]["splits"]["train"]["src"] = train_path
        template["val_dataset"]["splits"]["val"]["src"] = val_path
        # add extra large vaccum box for molecules
        # if dataset_name == str(UMATask.OMOL):
        #     template["train_dataset"]["a2g_args"]["molecule_cell_size"] = 1000.0
        os.makedirs(output_dir / DATA_YAML_DIR, exist_ok=True)
        with open(
            output_dir / REGRESSION_LABEL_TO_TASK_YAML[regression_tasks], "w"
        ) as yaml_file:
            yaml.dump(template, yaml_file, default_flow_style=False, sort_keys=False)

    uma_finetune_yaml = TEMPLATE_DIR / UMA_SM_FINETUNE_YAML
    with open(uma_finetune_yaml) as file:
        template_ft = yaml.safe_load(file)
        template_ft["base_model_name"] = base_model_name
        template_ft["defaults"][0]["data"] = REGRESSION_LABEL_TO_TASK_YAML[
            regression_tasks
        ].stem
        template_ft["train_dataset"]["dataset_configs"][dataset_name] = template_ft[
            "train_dataset"
        ]["dataset_configs"].pop("DATASET_NAME")
        template_ft["val_dataset"]["dataset_configs"][dataset_name] = template_ft[
            "val_dataset"
        ]["dataset_configs"].pop("DATASET_NAME")
    with open(output_dir / UMA_SM_FINETUNE_YAML, "w") as yaml_file:
        yaml.dump(template_ft, yaml_file, default_flow_style=False, sort_keys=False)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--train-dir",
        type=str,
        required=True,
        help="Directory of ASE atoms objects to convert for training.",
    )
    parser.add_argument(
        "--val-dir",
        type=str,
        required=True,
        help="Directory of ASE atoms objects to convert for validation.",
    )
    parser.add_argument(
        "--uma-task",
        type=str,
        required=True,
        choices=[t.value for t in UMATask],
        help="choose a uma task to finetune",
    )
    parser.add_argument(
        "--regression-tasks",
        type=str,
        choices=["e", "ef", "efs"],
        required=True,
        help="Choose to finetune based on regression task set (you must have the corresponding labels in your dataset), can be energy (e), energy+force (ef) or energy+force+stress(efs)",
    )
    parser.add_argument(
        "--base-model",
        type=str,
        default="uma-s-1",
        help="Name of base uma model",
    )
    parser.add_argument(
        "--output-dir",
        type=Path,
        required=True,
        help="Output directory to save required finetuning artifacts.",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        default=8,
        help="Number of parallel workers for processing files.",
    )
    args = parser.parse_args()
    assert not Path(
        args.output_dir
    ).exists(), f"{args.output_dir} can't already exist, please choose a different dir"

    # Launch processing for training data
    train_path = args.output_dir / "train"
    launch_processing(args.train_dir, train_path, args.num_workers)
    force_rms, linref_coeff = compute_normalizer_and_linear_reference(
        train_path, args.num_workers
    )
    # don't use force rms if doing energy only training
    if args.regression_tasks == "e":
        force_rms = 1.0
    val_path = args.output_dir / "val"
    launch_processing(args.val_dir, val_path, args.num_workers)

    create_yaml(
        train_path=str(train_path),
        val_path=str(val_path),
        force_rms=float(force_rms),
        linref_coeff=linref_coeff,
        output_dir=args.output_dir,
        dataset_name=args.uma_task,
        regression_tasks=args.regression_tasks,
        base_model_name=args.base_model,
    )
    logging.info(f"Generated dataset and data config yaml in {args.output_dir}")
    logging.info(
        f"To run finetuning, run fairchem -c {args.output_dir}/{UMA_SM_FINETUNE_YAML}"
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import torch
from torch import nn

if TYPE_CHECKING:
    from fairchem.core.models.escaip.configs import (
        GlobalConfigs,
        GraphNeuralNetworksConfigs,
        MolecularGraphConfigs,
        RegularizationConfigs,
    )
    from fairchem.core.models.escaip.custom_types import GraphAttentionData
from fairchem.core.models.escaip.utils.graph_utils import map_sender_receiver_feature
from fairchem.core.models.escaip.utils.nn_utils import Activation, get_linear


class BaseGraphNeuralNetworkLayer(nn.Module):
    """
    Base class for Graph Neural Network layers.
    Used in InputLayer and EfficientGraphAttention.
    """

    def __init__(
        self,
        global_cfg: GlobalConfigs,
        molecular_graph_cfg: MolecularGraphConfigs,
        gnn_cfg: GraphNeuralNetworksConfigs,
        reg_cfg: RegularizationConfigs,
    ):
        super().__init__()

        # Atomic number embeddings
        # ref: escn https://github.com/Open-Catalyst-Project/ocp/blob/main/ocpmodels/models/escn/escn.py#L823
        self.source_atomic_embedding = nn.Embedding(
            molecular_graph_cfg.max_num_elements, gnn_cfg.atom_embedding_size
        )
        self.target_atomic_embedding = nn.Embedding(
            molecular_graph_cfg.max_num_elements, gnn_cfg.atom_embedding_size
        )
        nn.init.uniform_(self.source_atomic_embedding.weight.data, -0.001, 0.001)
        nn.init.uniform_(self.target_atomic_embedding.weight.data, -0.001, 0.001)

        # Node direction embedding
        self.source_direction_embedding = get_linear(
            in_features=gnn_cfg.node_direction_expansion_size,
            out_features=gnn_cfg.node_direction_embedding_size,
            activation=Activation(global_cfg.activation),
            bias=True,
            dropout=reg_cfg.mlp_dropout,
        )
        self.target_direction_embedding = get_linear(
            in_features=gnn_cfg.node_direction_expansion_size,
            out_features=gnn_cfg.node_direction_embedding_size,
            activation=Activation(global_cfg.activation),
            bias=True,
            dropout=reg_cfg.mlp_dropout,
        )

        # Edge distance embedding
        self.edge_distance_embedding = get_linear(
            in_features=gnn_cfg.edge_distance_expansion_size,
            out_features=gnn_cfg.edge_distance_embedding_size,
            activation=Activation(global_cfg.activation),
            bias=True,
            dropout=reg_cfg.mlp_dropout,
        )

    def get_edge_linear(
        self,
        gnn_cfg: GraphNeuralNetworksConfigs,
        global_cfg: GlobalConfigs,
        reg_cfg: RegularizationConfigs,
    ):
        return get_linear(
            in_features=gnn_cfg.edge_distance_embedding_size
            + 2 * gnn_cfg.node_direction_embedding_size
            + 2 * gnn_cfg.atom_embedding_size,
            out_features=global_cfg.hidden_size,
            activation=Activation(global_cfg.activation),
            bias=True,
            dropout=reg_cfg.mlp_dropout,
        )

    def get_node_linear(
        self, global_cfg: GlobalConfigs, reg_cfg: RegularizationConfigs
    ):
        return get_linear(
            in_features=2 * global_cfg.hidden_size,
            out_features=global_cfg.hidden_size,
            activation=Activation(global_cfg.activation),
            bias=True,
            dropout=reg_cfg.mlp_dropout,
        )

    def get_edge_features(self, x: GraphAttentionData) -> torch.Tensor:
        # Atomic number embeddings (ref: escn)
        source_atomic_embedding = self.source_atomic_embedding(x.atomic_numbers)
        target_atomic_embedding = self.target_atomic_embedding(x.atomic_numbers)
        source_atomic_embedding, target_atomic_embedding = map_sender_receiver_feature(
            source_atomic_embedding, target_atomic_embedding, x.neighbor_list
        )

        # Node direction embedding
        source_direction_embedding = self.source_direction_embedding(
            x.node_direction_expansion
        )
        target_direction_embedding = self.target_direction_embedding(
            x.node_direction_expansion
        )
        source_direction_embedding, target_direction_embedding = (
            map_sender_receiver_feature(
                source_direction_embedding, target_direction_embedding, x.neighbor_list
            )
        )

        # Edge distance embedding
        edge_distance_embedding = self.edge_distance_embedding(
            x.edge_distance_expansion
        )

        # Concatenate edge features
        return torch.cat(
            [
                edge_distance_embedding,
                source_direction_embedding,
                source_atomic_embedding,
                target_direction_embedding,
                target_atomic_embedding,
            ],
            dim=-1,
        )

    def get_node_features(
        self, node_features: torch.Tensor, neighbor_list: torch.Tensor
    ) -> torch.Tensor:
        sender_feature, receiver_feature = map_sender_receiver_feature(
            node_features, node_features, neighbor_list
        )
        return torch.cat([sender_feature, receiver_feature], dim=-1)

    def aggregate(self, edge_features, neighbor_mask):
        return edge_features[:, 0]
        # neighbor_count = neighbor_mask.sum(dim=1, keepdim=True) + 1e-5
        # neighbor_count = neighbor_count.to(edge_features.dtype)
        # return (edge_features * neighbor_mask.unsqueeze(-1)).sum(dim=1) / neighbor_count

    def forward(self):
        raise NotImplementedError
